// EmEngine.cpp: implementation of the CEmEngine class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"

#include "Sockets/TaskProcessorSockets.h"
#include "Files/TaskProcessorFiles.h"
#include "Database/TaskProcessorDB.h"
#include "Other/TaskProcessorLogger.h"

#include "Other/EmWinNT.h"
#include "Other/EmMt.h"
#include "Other/ThreadPool.h"
#include "TaskProcessor.h"
#include "EmEngine.h"
#include "Sockets/TasksSockets.h"
#include "Sockets/TasksOpcodes.h"

#include "../Preferences.h"

#include "Data/Server.h"
//#include "Files/KnownFile.h"
//#include "Other/ed2k_filetype.h" included indirect over KnownFile.h

#include "XML/XML.h"

//////////////////////////////////////////////////////////////////////
// Constructor
CEmEngine::CEmEngine() :
	 m_hStop(NULL)
{
	ServerState.nConnState = SERVER_DISCONNECTED;
	ServerState.hSocket = NULL;
	ServerState.nAddr = 0;
	ServerState.nPort = 0;
	ServerState.nClientID = 0;

	m_dwThreadId = GetCurrentThreadId();
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// start
bool CEmEngine::Init()
{
/*	CServer stSrv;
	stSrv.ImportFromServerMet(*this, _T("./Db/server.met"));

	if(stSrv.ListStartByStatic(*this))
	{
		while(stSrv.ListGetNext())
		{
			int a = 1;
		}
		stSrv.ListFinish();
	}*/

//	Create an Event which can be asserted to stop the Engine. If successful...
	// Preferences
	m_pPreferences = new CPreferences;
	if (m_hStop = CreateEvent(NULL, TRUE, FALSE, NULL))
	{
		//	Try to initialize the Logger
		m_pLoggerProcessor = new CTaskProcessor_Logger;
		if (!m_pLoggerProcessor->Init())
		{
			ASSERT(FALSE);
			AddDebugLogLine("Failed to initialize logger processor");
			return false;
		}
		//	Try to initialize the Logger
		m_pDbProcessor = new CTaskProcessor_DB;
		if (!m_pDbProcessor->Init())
		{
			ASSERT(FALSE);
			AddDebugLogLine("Failed to initialize database processor");
			return false;
		}
		//	Try to initialize the Logger
		m_pFilesProcessor = new CTaskProcessor_Files;
		if (!m_pFilesProcessor->Init())
		{
			ASSERT(FALSE);
			AddDebugLogLine("Failed to initialize files processor");
			return false;
		}
		//	Initialize the Sockets TaskProcessor
		m_pSocketsProcessor = new CTaskProcessor_Sockets;
		if (m_pSocketsProcessor->Init())
		{
			//	Initialize the TCP "Engine".
			if (m_pSocketsProcessor->m_stTcp.Init())
			{
				// add relevant interfaces
				Sockets.m_stTcp.AddInterface(9090, T_CLIENT_XML);
				Sockets.m_stTcp.AddInterface(Prefs.GetPort(), T_CLIENT_PEER);
				//m_stTcp.AddInterface(80, T_CLIENT_WEB);
			}
			else
			{
				ASSERT(FALSE);
				AddDebugLogLine("Failed to initialize sockets processor");
				return false;
			}
		}
		else
		{
			ASSERT(FALSE);
			AddDebugLogLine("Failed to initialize sockets processor");
			return false;
		}
		m_pMainProcessor = new CTaskProcessor_Main;
		if (!m_pMainProcessor->Init())
		{
			ASSERT(FALSE);
			AddDebugLogLine("Failed to initialize main processor");
			return false;
		}
		//initial sort file types array
/*		if (!InitFileTypeArray())
		{
			ASSERT(FALSE);
			AddDebugLogLine("Failed to initialize filetype array");
			return false;
		}*/

		return true;
	}
	else
	{
		ASSERT(FALSE);
		AddDebugLogLine("CreateEvent failed");
		return false;
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// stop
void CEmEngine::Uninit()
{
	if (m_hStop)
	{
		if (!SetEvent(m_hStop))
		{
			ASSERT(FALSE);
			AddDebugLogLine("SetEvent failed");
		}

		m_pSocketsProcessor->m_stTcp.Uninit(false);
		m_pSocketsProcessor->Uninit();

		m_pMainProcessor->Uninit();
		delete m_pMainProcessor;

		m_pSocketsProcessor->m_stTcp.Uninit(); // final
		delete m_pSocketsProcessor;

		m_pLoggerProcessor->Uninit();
		delete m_pLoggerProcessor;

		m_pFilesProcessor->Uninit();
		delete m_pFilesProcessor;

		m_pDbProcessor->Uninit();
		delete m_pDbProcessor;

		delete m_pPreferences;

		if (!CloseHandle(m_hStop))
		{
			ASSERT(FALSE);
			AddDebugLogLine("CloseHandle failed");
		}
		m_hStop = NULL;
	}
}

bool CEmEngine::SendOpCode(SOCKET hSocket, const COpCode &stOpCode, CEmClient* pClient)
{
	if (pClient)
		m_pLoggerProcessor->Post(stOpCode, pClient->GetType(), pClient->m_nClientID, TRUE);
	return m_pSocketsProcessor->m_stTcp.AllocSend(hSocket, stOpCode);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// connecting to given server addr/port
void CEmEngine::ConnectToServer(CString sAddr/* = "" */, ULONG ulPort/* = 0 */)
{
	EMULE_TRY

	if (GetServerSocket())
	{
		DisconnectFromServer(GetServerSocket());
	}

//	AddLogLine(false, IDS_CONNECTINGTO, pServer->GetListName(), pServer->GetAddress(), pServer->GetPort());
	TRACE(_T("Connecting to server (%s:%d)\n"), sAddr, ulPort);

	CTask_LoginToServer	*pCompletionTask = new CTask_LoginToServer();
	CTask_Connect		*pTask = new CTask_Connect(sAddr, (USHORT) ulPort, T_CLIENT_SERVER, pCompletionTask);

	SetConnectionState(SERVER_CONNECTING);
	m_pSocketsProcessor->Push(pTask);

	EMULE_CATCH2
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// connect to any server
void CEmEngine::ConnectToAnyServer()
{
	ConnectToServer();
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// disconnect from connected server
void CEmEngine::DisconnectFromServer(SOCKET hPrevServer /* = NULL */)
{
	if(!hPrevServer)
	{
		if(GetServerSocket())
			DisconnectFromServer(GetServerSocket());
		else
			SetConnectionState(SERVER_DISCONNECTED);
		return;
	}

	CEmClient* pServer = m_pSocketsProcessor->Lookup(hPrevServer);
	if(!pServer)
	{
		ASSERT(FALSE);
		AddDebugLogLine(_T("lookup for server failed"));
		return;
	}
	CTask_KillClient* pTask = new CTask_KillClient(pServer);
	m_pSocketsProcessor->Push(pTask);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CEmEngine::SetConnectionState(short nConnectionState)
{
	ServerState.nConnState = nConnectionState;

//	Export connection state
	switch (nConnectionState)
	{
		case SERVER_DISCONNECTED:
		{
			ServerState.hSocket = NULL;
			ServerState.nAddr = 0;
			ServerState.nPort = 0;
			ServerState.nClientID = 0;
			TRACE("\nCONNECTION STATE CHANGE: DISCONNECTED\n");
			break;
		}
		case SERVER_CONNECTING:
		{
			TRACE("\nCONNECTION STATE CHANGE: CONNECTING\n");
			break;
		}
		case SERVER_WAITFORLOGIN:
		{
			TRACE("\nCONNECTION STATE CHANGE: WAITING FOR LOGIN\n");
			break;
		}
		case SERVER_CONNECTED:
		{
			TRACE("\nCONNECTION STATE CHANGE: CONNECTED\n");
			// send list of shared files
			CTask_SendSharedList* pTask = new CTask_SendSharedList();
			if(pTask)
				Sockets.Push(pTask);
			break;
		}
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CEmEngine::SetClientID(long cID)
{
	ServerState.nClientID = cID; 
}

//////////////////////////////////////////////////////////////////////
void CEmEngine::ConnectedTo(CEmClient_Server* pClient)
{
	ServerState.hSocket = pClient->m_hSocket;
	ServerState.nAddr = pClient->m_dwAddr;
	ServerState.nPort = pClient->m_uPort;
}

//////////////////////////////////////////////////////////////////////
bool CEmEngine::IsLocalServer(ULONG nAddr, USHORT uPort)
{
	if(!ServerState.IsConnected())
		return false;
	/*	if(m_stConnected.pServer)
	{
	if (m_stConnected.pServer->GetIP() == nAddr &&
	m_stConnected.pServer->GetPort() == uPort)
	return true;
	}*/
	return false;
}

//////////////////////////////////////////////////////////////////////
/*
#ifdef _DEBUG
static BYTE g_arrUserHash[16] = { 0x6c, 0x7c, 0x5d, 0x3b, 0x90, 0x0e, 0x11, 0x3f, 0xb1, 0x5c, 0x1c, 0x2e, 0xe8, 0x19, 0x6f, 0x59 };
#endif //_DEBUG

BYTE* CEmEngine::GetUserHash()
{
	//	return g_eMuleApp.m_pGlobPrefs->GetUserHash();
#ifdef _DEBUG
	return g_arrUserHash;
#else
	return NULL;
#endif
}

USHORT CEmEngine::GetPort()
{
	//	return g_eMuleApp.m_pGlobPrefs->GetPort();
	return 4662;
}

CString CEmEngine::GetUserNick()
{
	//	return g_eMuleApp.m_pGlobPrefs->GetUserNick();
	return _T("ePlus v2 debug (rus)(de)");
}

CString	CEmEngine::ConfigDirTemp()
{
	return _T("./temp");
}
*/

/*
// Contain all information about connecting/connected client
class CEmClientData
{
public:
	CPartFile*	GetFile() const { return m_pPartFile; }

private:
	CPartFile	*m_pPartFile;
};


//////////////////////////////////////////////////////////////////////
// Executed in GUI thread
void CEngineData::ImportServerState(CTask_ServerState* pTask)
{
	if(!pTask)
	{
		ASSERT(FALSE);
		return;
	}

	// should keep independent serverlist in the future
	if(!m_pServerList)
		m_pServerList = g_eMuleApp.m_pServerList;

	if(m_nClientID != pTask->m_nClientID)
	{
		m_nClientID = pTask->m_nClientID;
		AddLogLine(false, IDS_NEWCLIENTID, m_nClientID);
	}

	pState->pServer = m_pServerList->GetServerByIP(pState->nAddr, pState->nPort);

	if(pState->nConnState != pTask->m_stState.nConnState)
	{
		pState->nConnState = pTask->m_stState.nConnState;
		switch(pState->nConnState)
		{
		case SERVER_DISCONNECTED:
			if(m_stConnected.nConnState == SERVER_DISCONNECTED)
				g_eMuleApp.m_pdlgEmule->ShowConnectionState(false);
			break;
		case SERVER_CONNECTING:
			g_eMuleApp.m_pdlgEmule->ShowConnectionState(false);
			break;
		case SERVER_WAITFORLOGIN:
			if(m_stConnecting.pServer)
				AddLogLine(false, IDS_CONNECTEDTOREQ,
				m_stConnecting.pServer->GetListName(),
				m_stConnecting.pServer->GetFullIP(),
				m_stConnecting.pServer->GetPort());
			break;
		case SERVER_CONNECTED:
			if(m_stConnected.pServer)
			{
				AddLogLine(true, IDS_CONNECTEDTO, m_stConnected.pServer->GetListName());
				g_eMuleApp.m_pdlgEmule->ShowConnectionState(true, m_stConnected.pServer->GetListName());
				//				if(IsLowID())
				//					AddLogLine(false, GetResString(IDS_GOTLOWID));

				// Write online signature
				g_eMuleApp.OnlineSig();

				// Resume downloading files
				g_eMuleApp.m_pDownloadQueue->ResumeFiles();

				m_stEngine.SendSharedListToServer();
			}
			break;
		}
	}
}
*/
