// Tasks.cpp: implementation of the CTasks class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "OpCode.h"
#include "TasksOpcodes.h"
#include "../Files/TaskProcessorFiles.h"
#include "../../OtherFunctions.h"
#include "../../Preferences.h"
#include "../../SharedFileList.h"
#include "../../SafeFile.h"


OFFEREDFILE& OFFEREDFILE::operator = (const OFFEREDFILE& stObj)
{
	md4cpy(_Hash, stObj._Hash);
	_ClientID	= stObj._ClientID;
	_ClientPort	= stObj._ClientPort;
	_FileName	= stObj._FileName;
	_FileSize	= stObj._FileSize;
	_FileType	= stObj._FileType;

	return *this;
}


//////////////////////////////////////////////////////////////////////
// Client OpCodes
//////////////////////////////////////////////////////////////////////

void SendEmuleInfoAnswer(SOCKET hSocket, CEmClient_Peer* pClient)
{
	COpCode_EMULEINFOANSWER stMsg;
	stMsg._ClientVersion = CURRENT_VERSION_SHORT;
	stMsg._ProtocolVersion = EMULE_PROTOCOL;
	stMsg._Compression = 1;
	stMsg._UdpVersion = 4;
	stMsg._UdpPort = g_stEngine.Prefs.GetUDPPort();
	stMsg._SourceExchange = 0;//2;
	stMsg._Comments = 1;
	stMsg._ModPlus = CURRENT_PLUS_VERSION;
	stMsg._ExtendedRequest = 2;
	stMsg._ModVersion = _T(PLUS_VERSION_STR);
	stMsg._L2Hac = FILEREASKTIME;
	stMsg._Features = 0; // g_eMuleApp.m_pClientCreditList->CryptoAvailable() ? 3 : 0

	g_stEngine.SendOpCode(hSocket, stMsg, pClient);
}

//////////////////////////////////////////////////////////////////////
// HELLO opcode
bool COpCode_HELLO::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

//	CTask_HelloAnswer* pTask = new CTask_HelloAnswer(m_hSocket, pClient);
//	if(pTask)
//		g_stEngine.Sockets.Push(pTask);

	//	// here we should grab all user's info
	COpCode_HELLOANSWER stMsg;
	md4cpy(stMsg._UserHash, g_stEngine.Prefs.GetUserHash());
	stMsg._ClientAddr._IPAddr = g_stEngine.GetClientID();
	stMsg._ClientAddr._Port = g_stEngine.Prefs.GetPort();
	stMsg._UserName = g_stEngine.Prefs.GetUserNick();
	stMsg._ClientSoftVersion = EDONKEYVERSION;
	stMsg._UserUDPPort = g_stEngine.Prefs.GetUDPPort();
	stMsg._MiscOptions = 		
		(4						<< 4*6) |	// UDP version
		(1						<< 4*5) |	// Data compression version
		(0/*dwSupportSecIdent*/	<< 4*4) |	// Secure Ident
		(0/*2*/					<< 4*3) |	// Source Exchange
		(2						<< 4*2) |	// Ext. Requests
		(1						<< 4*1) |	// Comments
		(1/*dwNoViewSharedFiles*/<< 1*2);	// No View Shared files
	stMsg._ClientMuleVersion = 
		(PLUS_COMPATIBLECLIENTID << 24) | 
		((CURRENT_PLUS_VERSION & 0x7F00) << 9) |
		((CURRENT_PLUS_VERSION & 0xF0) << 6) | 
		((CURRENT_PLUS_VERSION & 0x7) << 7);
	/*	if(g_stEngine.ServerState.IsConnecting() || g_stEngine.ServerState.IsConnected())
	{
	stMsg._ServerAddr._IPAddr = g_stEngine.ServerState.nAddr;
	stMsg._ServerAddr._Port = g_stEngine.ServerState.nPort;
	}
	else
	{
	stMsg._ServerAddr._IPAddr = 0;
	stMsg._ServerAddr._Port = 0;
	}*/

	g_stEngine.SendOpCode(m_hSocket, stMsg, pClient);

	if (1/*emule protocol supported*/)
	{
		SendEmuleInfoAnswer(m_hSocket, pClient);
	}

	return true;
}

//////////////////////////////////////////////////////////////////////
// HELLO ANSWER
bool COpCode_HELLOANSWER::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

bool COpCode_EMULEINFO::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	SendEmuleInfoAnswer(m_hSocket, pClient);

	return true;
}

bool COpCode_EMULEINFOANSWER::ProcessForClient(CEmClient_Peer* pClient)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Request filename
bool COpCode_REQUESTFILENAME::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	CTask_SendRequestedFileName* pTask = new CTask_SendRequestedFileName(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Request filename answer
bool COpCode_REQFILENAMEANSWER::ProcessForClient(CEmClient_Peer* pClient)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Start file upload
bool COpCode_STARTUPLOADREQ::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	CTask_UploadReqResult* pTask = new CTask_UploadReqResult(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Accept upload request
bool COpCode_ACCEPTUPLOADREQ::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// No such file
bool COpCode_FILEREQANSNOFIL::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Request parts
bool COpCode_REQUESTPARTS::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	CTask_SendBlock* 
	pTask = new CTask_SendBlock(_Hash, pClient, _Start1, _End1);
	if(pTask)
		g_stEngine.Files.Push(pTask);
	pTask = new CTask_SendBlock(_Hash, pClient, _Start2, _End2);
	if(pTask)
		g_stEngine.Files.Push(pTask);
	pTask = new CTask_SendBlock(_Hash, pClient, _Start3, _End3);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Sending file part
bool COpCode_SENDINGPART::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Get file status
bool COpCode_SETREQFILEID::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	CTask_SendFileStatus* pTask = new CTask_SendFileStatus(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// File status
bool COpCode_FILESTATUS::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Hash request
bool COpCode_HASHSETREQUEST::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	CTask_SendHashsets* pTask = new CTask_SendHashsets(_Hash, pClient);
	if(pTask)
		g_stEngine.Files.Push(pTask);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Hash request answer
bool COpCode_HASHSETANSWER::ProcessForClient(CEmClient_Peer*)
{
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Offer files
bool COpCode_OFFERFILES::ProcessForClient(CEmClient_Peer*)
{
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
// Server OpCodes
//////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SERVER MESSAGE
bool COpCode_SERVERMESSAGE::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	CEmClient_Server* pSrv = reinterpret_cast<CEmClient_Server*>(pClient);

	// Increase message count
	m_nMsgCount = pSrv->m_nMsgCount++;

	TRACE("MESSAGE FROM SERVER: %s\n", _Msg);

//	m_Server._Addr = pSrv->m_dwAddr;
//	m_Server._Port = pSrv->m_uPort;

//	stEngine.PushToUI(this);

	return false; // Always return false when going to GUI, otherwise this object will be destroyed!
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SERVER MESSAGE in GUI
/*bool COpCode_SERVERMESSAGE::ProcessForUI()
{
	// Print message
	g_eMuleApp.m_pdlgEmule->AddServerMessageLine(_Msg);

	// Check for server version
	if ((m_nMsgCount < 2) && (_Msg.Left(14).CompareNoCase(_T("server version")) == 0))
	{
		CString strVer = _Msg.Mid(14);
		strVer.Trim();
		strVer = strVer.Left(64); // Truncate string to avoid misuse by servers in showing ads
		int n = strVer.FindOneOf("\x09\x0A\x0D");
		if (n >= 0)
			strVer = strVer.Left(n);
		CServer* pServer = g_eMuleApp.m_pServerList->GetServerByIP(m_Server._Addr, m_Server._Port);
		if (pServer)
			pServer->SetVersion(strVer);
	}

	// Check for dynamic IP
	int nDynIP = _Msg.Find(_T("[emDynIP: "));
	int nCloseBracket = _Msg.Find(_T("]"));
	if (nDynIP != -1 && nCloseBracket != -1 && nDynIP < nCloseBracket)
	{
		CString sDynIP = _Msg.Mid(nDynIP + 10, nCloseBracket - (nDynIP + 10));
		sDynIP.Trim(_T(" "));
		if (sDynIP.GetLength() && sDynIP.GetLength() < 51)
		{
			CServer* pServer = g_eMuleApp.m_pServerList->GetServerByIP(m_Server._Addr, m_Server._Port);
			if (pServer)
			{
				pServer->SetDynIP(sDynIP.GetBuffer());
				g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
			}
		}
	}

	return true;
}*/
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// ID CHANGE
bool COpCode_IDCHANGE::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// TODO: Smart ID checks, notifiers

	// Disconnect from previous connected server, if any
	if(g_stEngine.GetServerSocket() != pClient->m_hSocket &&
		g_stEngine.GetServerSocket())
		g_stEngine.DisconnectFromServer(g_stEngine.GetServerSocket());

	// Set server connection info
	g_stEngine.ConnectedTo(reinterpret_cast<CEmClient_Server*>(pClient));
	g_stEngine.SetClientID(_NewClientID);
	g_stEngine.SetConnectionState(SERVER_CONNECTED);

/*
	if (g_eMuleApp.m_pGlobPrefs.GetAddServersFromConn())
	{
		Packet* packet = new Packet(OP_GETSERVERLIST);
		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadServer(packet->size);
		SendPacket(packet,true);
	}
	if(g_eMuleApp.m_pGlobPrefs.RestartWaiting())
		g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlDownloadList.RestartWaitingDownloads();
 */

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SEARCH RESULT
bool COpCode_SEARCHRESULT::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FOUND SOURCES
bool COpCode_FOUNDSOURCES::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SERVER STATUS
bool COpCode_SERVERSTATUS::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	// Set address and port
//	m_Server._Addr = pClient->m_dwAddr;
//	m_Server._Port = pClient->m_uPort;

//	stEngine.PushToUI(this);

	TRACE("SERVER STATUS: %u users, %u files\n", _NumberOfUsers, _NumberOfFiles);

	return false; // Always return false when going to GUI, otherwise this object will be destroyed!
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SERVER STATUS in GUI
/*bool COpCode_SERVERSTATUS::ProcessForUI()
{
	CServer* pServer = g_eMuleApp.m_pServerList->GetServerByIP(m_Server._Addr, m_Server._Port);
	if (pServer)
	{
		pServer->SetUserCount(_NumberOfUsers); 
		pServer->SetFileCount(_NumberOfFiles);
		g_eMuleApp.m_pdlgEmule->ShowUserCount(_NumberOfUsers, _NumberOfFiles);
		g_eMuleApp.m_pdlgEmule->m_wndServer.m_ctlServerList.RefreshServer(*pServer);
	}

	return true;
}*/
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SERVER IDENTification
bool COpCode_SERVERIDENT::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// SERVER LIST
bool COpCode_SERVERLIST::ProcessForClient(CEmClient_Peer* pClient)
{
	ASSERT(pClient); // this opcode is processed only in the main context

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CALLBACKREQUEST is a Client => Server message and should never be
//		received by a client.
bool COpCode_CALLBACKREQUEST::ProcessForClient(CEmClient_Peer *pClient)
{
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CALLBACKFAIL (if it is ever received by a client) is ignored by
//		eMulePlus.
bool COpCode_CALLBACKFAIL::ProcessForClient(CEmClient_Peer *pClient)
{
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// CALLBACK REQUESTED
bool COpCode_CALLBACKREQUESTED::ProcessForClient(CEmClient_Peer* )
{
	CTask_SayHelloToPeer   *pSayHelloTask = new CTask_SayHelloToPeer();
	CTask_Connect		   *pTask = new CTask_Connect(_ClientAddr._IPAddr, _ClientAddr._Port, T_CLIENT_PEER, pSayHelloTask);

	g_stEngine.Sockets.Push(pTask);

	AddDebugLogLine("Callback requested");

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// LOGIN REQUEST
bool COpCode_LOGINREQUEST::ProcessForClient(CEmClient_Peer* )
{
	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_LoginToServer::Process()
{
	//	Reset message count
	//	m_pServer->m_nMsgCount = 0;
	if (m_pServer == NULL)
	{
		//	Report error...
		return true;
	}
	if (m_iMessage == FD_CONNECT)
	{
		//
		//	We've just connected. Make sure we're prepared to handle disconnect.
		CTask_DisconnectServer		*pOnDisconnectTask = new CTask_DisconnectServer();

		pOnDisconnectTask->m_pServer = m_pServer;
		m_pServer->m_pOnCompletionTask = pOnDisconnectTask;

		//	Disconnect from previous connected server, if any
		if ( g_stEngine.GetServerSocket() != m_pServer->m_hSocket
			&& g_stEngine.GetServerSocket() != 0 )
		{
			g_stEngine.DisconnectFromServer(g_stEngine.GetServerSocket());
		}

		//	Set connection info
		g_stEngine.ConnectedTo(m_pServer);
		g_stEngine.SetConnectionState(SERVER_WAITFORLOGIN);

		COpCode_LOGINREQUEST		stMsg;

		md4cpy(stMsg._UserHash, g_stEngine.Prefs.GetUserHash());
		stMsg._ClientAddr._IPAddr = g_stEngine.GetClientID();
		stMsg._ClientAddr._Port = g_stEngine.Prefs.GetPort();
		stMsg._UserName = g_stEngine.Prefs.GetUserNick();
		stMsg._ClientSoftVersion = EDONKEYVERSION;
		stMsg._UserPort = g_stEngine.Prefs.GetPort();
		stMsg._Flags = SRVCAP_ZLIB | SRVCAP_AUXPORT | SRVCAP_NEWTAGS;
		stMsg._eMuleVersion = (PLUS_COMPATIBLECLIENTID << 24) | ((CURRENT_PLUS_VERSION & 0x7F00) << 9) |
			((CURRENT_PLUS_VERSION & 0xF0) << 6) | ((CURRENT_PLUS_VERSION & 0x7) << 7);

		g_stEngine.SendOpCode(m_pServer->m_hSocket, stMsg, m_pServer);
	}
	else
	{
		//	We're not expecting this.
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CTask_DisconnectServer::Process()
{
	if (m_pServer == NULL)
		return true;

	// (!) at any reason, can be even if we tried to connect and failed

	if (m_iMessage == FD_CLOSE)
	{
		// Set connection info
		if (g_stEngine.GetServerSocket() == m_pServer->m_hSocket)
		{
			g_stEngine.SetConnectionState(SERVER_DISCONNECTED);
		}
		else
		{
			// removed if, because we can get 'disconnected' even if we didn't
			// succeeded to connect yet, thus GetConnectingServer() won't return
			// correct value
			// if(rEngine.GetConnectingServer() == m_hSocket)
			g_stEngine.SetConnectionState(SERVER_DISCONNECTED);
			// TODO: mark somewhere that server is dead
		}
	}
	else
	{
		//	We're not expecting this.
	}

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// (!) can be both we connected to him or he connected to us
// sending HELLO opcode.
//	MOREVIT: If I'm understanding this correctly, you're saying that
//		when another client connects to us to say HELLO, we're going
//		a HELLO back. What's the point in that? Once the HELLOANSWER
//		is received, both sides have all the information they need
//		about the other.
bool CTask_SayHelloToPeer::Process()
{
	if (m_pPeer == NULL)
		return true;

	if (m_iMessage == FD_CONNECT && !m_pPeer->m_bFromOutside)
	{
		COpCode_HELLO		stMsg;

		md4cpy(stMsg._UserHash, g_stEngine.Prefs.GetUserHash());
		stMsg._ClientAddr._IPAddr = g_stEngine.GetClientID();
		stMsg._ClientAddr._Port = g_stEngine.Prefs.GetPort();
		stMsg._UserName = g_stEngine.Prefs.GetUserNick();
		stMsg._ClientSoftVersion = EDONKEYVERSION;
		stMsg._UserPort = g_stEngine.Prefs.GetPort();

		if (g_stEngine.ServerState.IsConnected())
		{
			stMsg._ServerAddr._IPAddr = g_stEngine.ServerState.nAddr;
			stMsg._ServerAddr._Port = g_stEngine.ServerState.nPort;
		}
		else
		{
			stMsg._ServerAddr._IPAddr = 0;
			stMsg._ServerAddr._Port = 0;
		}

		g_stEngine.SendOpCode(m_pPeer->m_hSocket, stMsg, m_pPeer);
	}
	else
	{
	}

	return true;
}

enum
{
	TASK_START	= 0,
	TASK_READ	= 1,
	TASK_SEND	= 2
};

//////////////////////////////////////////////////////////////////////
// Send list of shared files
CTask_SendSharedList::CTask_SendSharedList()
{
	m_nState = TASK_START;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendSharedList::Process()
{
	if(!g_stEngine.ServerState.IsConnected())
		return true;

	switch(m_nState)
	{
	case TASK_START:
		m_nState = TASK_READ;
		g_stEngine.Files.Push(this);
		return false;
	case TASK_READ:
		if(g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
			return true;
		for(POSITION pos = g_stEngine.Files.SharedFiles.m_mapSharedFiles.GetStartPosition(); pos != NULL; )
		{
			CCKey		bufKey;
			CKnownFile*	pKnownFile = NULL;
			g_stEngine.Files.SharedFiles.m_mapSharedFiles.GetNextAssoc(pos, bufKey, pKnownFile);
			if(pKnownFile)
			{
				OFFEREDFILE stFile;
				md4cpy(stFile._Hash, pKnownFile->GetFileHash());

				uint32		dwClientID = 0;
				uint16		uClientPort = 0;

				if (false) // pServer && (pServer->GetTCPFlags() & SRV_TCPFLG_COMPRESSION)
				{
					if (pKnownFile->IsPartFile())
					{
						stFile._ClientID	= 0xFCFCFCFC;
						stFile._ClientPort	= 0xFCFC;
					}
					else
					{
						// Publishing a complete files
						stFile._ClientID	= 0xFBFBFBFB;
						stFile._ClientPort	= 0xFBFB;
					}
				}
				else
				{
					//	If connected and highID, set client ID and port
					if(true) // m_pServerConnect->IsConnected() && !m_pServerConnect->IsLowID()
					{
						stFile._ClientID	= g_stEngine.GetClientID();
						stFile._ClientPort	= g_stEngine.Prefs.GetPort();
					}
				}
				stFile._FileName = pKnownFile->GetFileName();
				stFile._FileSize = pKnownFile->GetFileSize();
				stFile._FileType = pKnownFile->GetFileTypeString();

				//	There's no need to send FT_FILEFORMAT (file extension without ".") here as a server takes it from the file name
				/*if ( ((pServer != NULL) && (pServer->GetTCPFlags() & SRV_TCPFLG_NEWTAGS)) ||
					( (pClient != NULL) && pClient->IsEmuleClient() &&
					( ((pClient->GetClientSoft() == SO_EMULE) && (pClient->GetVersion() >= FORM_CLIENT_VER(0, 42, 7))) ||
					((pClient->GetClientSoft() == SO_PLUS) && (pClient->GetVersion() > FORM_CLIENT_VER(1, 1, 0))) ) ) )
				*/
				m_Files.Add(stFile);
			}
		}
		m_nState = TASK_SEND;
		g_stEngine.Sockets.Push(this);
		return false;
	case TASK_SEND:
		if(m_Files.GetCount() > 0)
		{
			COpCode_OFFERFILES stMsg;
			for(int i = 0; i < m_Files.GetCount(); i++)
			{
				COpCode_OFFERFILES::CStruct_Files stFile;
				md4cpy(stFile._Hash, m_Files[i]._Hash);
				stFile._ClientAddr._IPAddr	= m_Files[i]._ClientID;
				stFile._ClientAddr._Port	= m_Files[i]._ClientPort;
				stFile._FileName			= m_Files[i]._FileName;
				stFile._FileSize			= m_Files[i]._FileSize;
				stFile._FileType			= m_Files[i]._FileType;
				stMsg._Files.push_back(stFile);
			}

			SOCKET hServerSocket = g_stEngine.ServerState.hSocket;
			CEmClient* pClient = g_stEngine.Sockets.Lookup(hServerSocket);
			if(pClient)
				g_stEngine.SendOpCode(hServerSocket, stMsg, pClient);
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send requested filename
CTask_SendRequestedFileName::CTask_SendRequestedFileName(BYTE Hash[16], CEmClient* pClient)
{
	m_nState = TASK_READ;
	md4cpy(m_Hash, Hash);
	m_pClient = pClient;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendRequestedFileName::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash);
			if(pFile)
			{
				m_strFileName	= pFile->GetFileName();
				m_nState = TASK_SEND;
				g_stEngine.Sockets.Push(this);
				return false;
			}
		}
		return true;
	case TASK_SEND:
		if(!m_strFileName.IsEmpty() && m_pClient)
		{
			COpCode_REQFILENAMEANSWER stMsg;
			md4cpy(stMsg._Hash, m_Hash);
			stMsg._FileName	= m_strFileName;
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Start file upload
CTask_UploadReqResult::CTask_UploadReqResult(BYTE Hash[16], CEmClient* pClient)
{
	m_nState = TASK_READ;
	md4cpy(m_Hash, Hash);
	m_pClient = pClient;
}

//////////////////////////////////////////////////////////////////////
bool CTask_UploadReqResult::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		m_bFound = false;
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash);
			if(pFile)
				m_bFound = true;
		}
		m_nState = TASK_SEND;
		g_stEngine.Sockets.Push(this);
		return false;
	case TASK_SEND:
		if(m_pClient)
		{
			if(m_bFound)
			{
				COpCode_ACCEPTUPLOADREQ stMsg;
				g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			}
			else
			{
				COpCode_FILEREQANSNOFIL stMsg;
				md4cpy(stMsg._Hash, m_Hash);
				g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			}
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send requested block
CTask_SendBlock::CTask_SendBlock(BYTE Hash[16], CEmClient* pClient, DWORD dwStart, DWORD dwEnd)
{
	m_nState = TASK_READ;
	md4cpy(m_Hash, Hash);
	m_pClient = pClient;
	m_dwStart = dwStart;
	m_dwEnd = dwEnd;
	m_dwToGo = 0;
	m_pFileData = NULL;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendBlock::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash);
			if(pFile && m_dwStart < pFile->GetFileSize())
			{
				DWORD dwPart = m_dwStart / PARTSIZE;
				DWORD dwStart = dwPart * PARTSIZE;
				DWORD dwEnd = ((pFile->GetFileSize() - dwStart) > PARTSIZE) ? (dwStart + PARTSIZE) : pFile->GetFileSize();
				//	Everyone is limited to a single chunk.
				//	Check that m_dwStartOffset and m_dwEndOffset are in the same chunk and
				//	make sure we don't pass the end of the file.
				if((dwPart != ((m_dwEnd - 1) / PARTSIZE)) || (m_dwEnd > dwEnd))
				{
					//	m_dwEnd goes into the next chunk or is beyond the file end.
					//	Set it to the end of the chunk that m_dwStartOffset is in.
					m_dwEnd = dwEnd;
				}
				//	This can't be a wrapped around request, since it has been limited to a single chunk.
				m_dwToGo = m_dwEnd - m_dwStart;
				//	Create a buffer for file data before we start to work with the file
				m_pFileData = new BYTE[m_dwToGo + 500];
				int	iRc = pFile->ReadFileForUpload(m_dwStart, m_dwToGo, m_pFileData);
				if (iRc >= 0)
				{
					m_nState = TASK_SEND;
					g_stEngine.Sockets.Push(this);
					return false;
				}
				else
					delete[] m_pFileData;
			}
		}
		return true;
	case TASK_SEND:
		if(m_pClient && m_pFileData)
		{
			COpCode_SENDINGPART stMsg;
			md4cpy(stMsg._Hash, m_Hash);
			stMsg._StartData = m_dwStart;
			stMsg._EndData = m_dwEnd;
			stMsg._Data = new BYTE[m_dwToGo];
			CopyMemory(stMsg._Data, m_pFileData, m_dwToGo);
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			delete[] m_pFileData;
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send file status
CTask_SendFileStatus::CTask_SendFileStatus(BYTE Hash[16], CEmClient* pClient)
{
	m_nState = TASK_READ;
	md4cpy(m_Hash, Hash);
	m_pClient = pClient;
	m_pFileStatus = NULL;
	m_nStatusSize = 0;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendFileStatus::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash);
			if(pFile)
			{
//				CSafeMemFile packetStream(16);
//				((CPartFile*)pFile)->WritePartStatus(&packetStream);
//				m_nStatusSize = packetStream.GetLength();
//				m_pFileStatus = new BYTE[m_nStatusSize];
//				packetStream.Write(m_pFileStatus, m_nStatusSize);

				m_nStatusSize = pFile->GetED2KPartCount();
				m_pFileStatus = new BYTE[m_nStatusSize];
				for(int i = 0; i < m_nStatusSize; i++)
					m_pFileStatus[i] = 1;

				m_nState = TASK_SEND;
				g_stEngine.Sockets.Push(this);
				return false;
			}
		}
		return true;
	case TASK_SEND:
		if(m_pClient && m_pFileStatus)
		{
			COpCode_FILESTATUS stMsg;
			md4cpy(stMsg._Hash, m_Hash);
			for(int i = 0; i < m_nStatusSize; i++)
				stMsg._FileStatus.push_back(m_pFileStatus[i]);
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			delete[] m_pFileStatus;
		}
	}
	return true;
}

//////////////////////////////////////////////////////////////////////
// Send file hashsets
CTask_SendHashsets::CTask_SendHashsets(BYTE Hash[16], CEmClient* pClient)
{
	m_nState = TASK_READ;
	md4cpy(m_Hash, Hash);
	m_pClient = pClient;
	m_pFileHashes = NULL;
	m_nHashCount = 0;
}

//////////////////////////////////////////////////////////////////////
bool CTask_SendHashsets::Process()
{
	switch(m_nState)
	{
	case TASK_READ:
		if(!g_stEngine.Files.SharedFiles.m_mapSharedFiles.IsEmpty())
		{
			CKnownFile* pFile = g_stEngine.Files.SharedFiles.GetFileByID(m_Hash);
			if(pFile && pFile->GetHashSet() != NULL)
			{
				m_nHashCount = pFile->GetHashCount();
				m_pFileHashes = new HASH[m_nHashCount];
				CopyMemory(m_pFileHashes, pFile->GetHashSet(), sizeof(HASH)*m_nHashCount);

				m_nState = TASK_SEND;
				g_stEngine.Sockets.Push(this);
				return false;
			}
		}
		return true;
	case TASK_SEND:
		if(m_pClient && m_pFileHashes)
		{
			COpCode_HASHSETANSWER stMsg;
			md4cpy(stMsg._Hash, m_Hash);
			for(int i = 0; i < m_nHashCount; i++)
				stMsg._Hashsets.push_back(m_pFileHashes[i]);
			g_stEngine.SendOpCode(m_pClient->m_hSocket, stMsg, m_pClient);
			delete[] m_pFileHashes;
		}
	}
	return true;
}
