// parts of this file are based on work from pan One (http://home-3.tiscali.nl/~meost/pms/)

//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "KnownFile.h"
#include "opcodes.h"
#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifndef NEW_SOCKETS_ENGINE
	#include "emule.h"
#endif //NEW_SOCKETS_ENGINE
#include "ini2.h"
#include "QArray.h"
#include "otherfunctions.h"
#ifndef NEW_SOCKETS_ENGINE
	#include "Jumpstarter.h"
#endif //NEW_SOCKETS_ENGINE

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CAbstractFile::CreateED2kLink() const
{
	CString		strLink;

	strLink.Format( _T("ed2k://|file|%s|%u|%s|/"),
#ifndef NEW_SOCKETS_ENGINE
					g_eMuleApp.StripInvalidFilenameChars(GetFileName(), false),	// spaces to dots
#else
					GetFileName(),
#endif //NEW_SOCKETS_ENGINE
					GetFileSize(),
					HashToString(GetFileHash()) );

	return strLink;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CAbstractFile::CreateED2kSourceLink() const
{
	CString		strLink = CreateED2kLink();
#ifdef OLD_SOCKETS_ENABLED

	if (g_eMuleApp.m_pServerConnect->IsConnected() && !g_eMuleApp.m_pServerConnect->IsLowID())
	{
		uint32			dwID = g_eMuleApp.m_pServerConnect->GetClientID();

		strLink.AppendFormat( _T("|sources,%i.%i.%i.%i:%i|/"),
								static_cast<BYTE>(dwID),
								static_cast<BYTE>(dwID>>8),
								static_cast<BYTE>(dwID>>16),
								static_cast<BYTE>(dwID>>24),
								g_eMuleApp.m_pGlobPrefs->GetListenPort() );
	}
#endif //OLD_SOCKETS_ENABLED

	return strLink;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CAbstractFile::CreateHTMLED2kLink() const
{
	return _T("<a href=\"") + CreateED2kLink() + _T("\">") + 
#ifndef NEW_SOCKETS_ENGINE
		g_eMuleApp.StripInvalidFilenameChars(GetFileName(), true) 
#else
		GetFileName()
#endif //NEW_SOCKETS_ENGINE
		+ _T("</a>");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CFileStatistic::CFileStatistic()
{
	partTraffic = NULL;
	partTrafficSession = NULL;

	blockTraffic = NULL;
	blockTrafficSession = NULL;

	partAccepted = NULL;
	partAcceptedSession = NULL;

	completeReleases = 0.0;
	m_iNumRequested = m_qwNumTransferred = m_iNumAccepted = alltimerequested= alltimetransferred = alltimeaccepted = 0;
	fileParent	=	NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CFileStatistic::~CFileStatistic()
{
	EMULE_TRY

	if (partTraffic)			delete[] partTraffic;			partTraffic=NULL;
	if (partTrafficSession)		delete[] partTrafficSession;	partTrafficSession=NULL;

	if (partAccepted)			delete[] partAccepted;			partAccepted=NULL;
	if (partAcceptedSession)	delete[] partAcceptedSession;	partAcceptedSession=NULL;

	if (blockTraffic)			delete[] blockTraffic;			blockTraffic=NULL;
	if (blockTrafficSession)	delete[] blockTrafficSession;	blockTrafficSession=NULL;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::AddRequest()
{
	EMULE_TRY

	m_iNumRequested++;
	alltimerequested++;
#ifndef NEW_SOCKETS_ENGINE
	g_eMuleApp.m_pKnownFilesList->m_iNumRequested++;
	if (::IsWindow(g_eMuleApp.m_pdlgEmule->m_wndSharedFiles.m_hWnd))	//fix crash at exit [TwoBottle Mod]
	{
		fileParent->UpdateSharedFileDisplay();
	}
#endif //NEW_SOCKETS_ENGINE
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::AddAccepted()
{
	EMULE_TRY

	m_iNumAccepted++;
	alltimeaccepted++;
#ifndef NEW_SOCKETS_ENGINE
	g_eMuleApp.m_pKnownFilesList->m_iNumAccepted++;
#endif //NEW_SOCKETS_ENGINE
	fileParent->UpdateSharedFileDisplay();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::AddTransferred(uint32 bytes)
{
	EMULE_TRY

	m_qwNumTransferred += bytes;
	alltimetransferred += bytes;
#ifndef NEW_SOCKETS_ENGINE
	g_eMuleApp.m_pKnownFilesList->m_iNumTransferred += bytes;
#endif //NEW_SOCKETS_ENGINE
	fileParent->UpdateSharedFileDisplay();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CKnownFile::CKnownFile() : m_strComment(_T(""))
{
	m_timetLastWriteDate = 0;
	m_bytePriority = PR_NORMAL;

#ifndef NEW_SOCKETS_ENGINE
	m_bAutoPriority = g_eMuleApp.m_pGlobPrefs->IsUAPEnabled();
#endif //NEW_SOCKETS_ENGINE

	m_bytePermissions = PERM_ALL;
	statistic.fileParent = this;
	m_bCommentLoaded = false;
	m_byteMoviePreviewMode = 0;
	m_eRating = PF_RATING_NONE;
	m_nCompleteSourcesCount = 0;
	m_nCompleteSourcesCountLo = 0;
	m_nCompleteSourcesCountHi = 0;
	m_nCompleteSourcesTime = 0;

	m_sharedFile = false;
	md4clr(m_fileHash);

	m_Jumpstarter = NULL;
	m_pPartsHashSet = NULL;
	m_dblSizeRatio = 1.0;
	m_bIsCompressedTransferAllowed = true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CKnownFile::~CKnownFile()
{
	EMULE_TRY

	CSingleLock Lock(&m_mutexHashList,TRUE);
	if (m_pPartsHashSet != NULL)
		delete m_pPartsHashSet;
	Lock.Unlock();

	for (int i = 0; i < m_tagArray.GetSize(); i++)
		safe_delete(m_tagArray[i]);

#ifndef NEW_SOCKETS_ENGINE
	if (m_Jumpstarter)
		delete m_Jumpstarter;
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::CalculateCompleteSources()
{
	EMULE_TRY

	uint32	dwCurTick;

#ifndef NEW_SOCKETS_ENGINE
	if (!IsPartFile() && (((dwCurTick = ::GetTickCount()) - m_nCompleteSourcesTime) >= 2*60*1000))	// 2 min.
	{
		CQArray<uint16,uint16> count;
		CTypedPtrList<CPtrList, CUpDownClient*> clientuploadlist;
		m_nCompleteSourcesCount= m_nCompleteSourcesCountLo= m_nCompleteSourcesCountHi= 0;
		g_eMuleApp.m_pClientList->GetClientListByFileID(&clientuploadlist, GetFileHash());
		count.SetSize(0, clientuploadlist.GetSize());
		for (POSITION pos= clientuploadlist.GetHeadPosition(); pos != NULL; )
		{
			CUpDownClient *cur_src= clientuploadlist.GetNext(pos);
			uint16 cur_count= cur_src->GetUpCompleteSourcesCount();
			if ((cur_count) && ((dwCurTick - cur_src->GetUpCompleteSourcesTime()) < 30*60*1000))	// 30 min.
				count.Add(cur_count);
		}

		uint32 n= count.GetSize();
		if (n > 0)
		{
			count.QuickSort();

		//	Calculate range
			uint32 i= n >> 1;		// (n / 2)
			uint32 j= (n * 3) >> 2;	// (n * 3) / 4
			uint32 k= (n * 7) >> 3;	// (n * 7) / 8
			if (n < 5)
			{
				m_nCompleteSourcesCount= count.GetAt(i);
				m_nCompleteSourcesCountLo= 0;
				m_nCompleteSourcesCountHi= m_nCompleteSourcesCount;
			}
			else if (n < 10)
			{
				m_nCompleteSourcesCount= count.GetAt(i);
				m_nCompleteSourcesCountLo= count.GetAt(i - 1);
				m_nCompleteSourcesCountHi= count.GetAt(i + 1);
			}
			else if (n < 20)
			{
				m_nCompleteSourcesCount= count.GetAt(i);
				m_nCompleteSourcesCountLo= count.GetAt(i);
				m_nCompleteSourcesCountHi= count.GetAt(j);
			}
			else
			{
				m_nCompleteSourcesCount= count.GetAt(j);
				m_nCompleteSourcesCountLo= m_nCompleteSourcesCount;
				m_nCompleteSourcesCountHi= count.GetAt(k);
			}
		}

		m_nCompleteSourcesTime = dwCurTick;
	}

	UpdateUploadAutoPriority();
	UpdateSharedFileDisplay();
#endif //NEW_SOCKETS_ENGINE
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::UpdateSharedFileDisplay()
{
#ifndef NEW_SOCKETS_ENGINE
	if (g_eMuleApp.m_pdlgEmule->m_wndSharedFiles.m_ctlSharedFilesList.m_hWnd != NULL)
	{
		g_eMuleApp.m_pdlgEmule->m_wndSharedFiles.m_ctlSharedFilesList.PostMessage( WM_SFL_UPDATEITEM,
																				static_cast<WPARAM>(g_eMuleApp.m_pdlgEmule->m_wndSharedFiles.IsWindowVisible()),
																				reinterpret_cast<LPARAM>(dynamic_cast<CKnownFile*>(this)) );
	}
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::SetFilePath(LPCTSTR pszFilePath)
{
	m_strFilePath = pszFilePath;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::SetFileSize(uint32 dwFileSize)
{
	CAbstractFile::SetFileSize(dwFileSize);

	// Examples of parthashs, hashsets and filehashs for different filesizes
	// according the ed2k protocol
	//----------------------------------------------------------------------
	//
	//File size: 3 bytes
	//File hash: 2D55E87D0E21F49B9AD25F98531F3724
	//Nr. hashs: 0
	//
	//
	//File size: 1*PARTSIZE
	//File hash: A72CA8DF7F07154E217C236C89C17619
	//Nr. hashs: 2
	//Hash[  0]: 4891ED2E5C9C49F442145A3A5F608299
	//Hash[  1]: 31D6CFE0D16AE931B73C59D7E0C089C0	*special part hash*
	//
	//
	//File size: 1*PARTSIZE + 1 byte
	//File hash: 2F620AE9D462CBB6A59FE8401D2B3D23
	//Nr. hashs: 2
	//Hash[  0]: 121795F0BEDE02DDC7C5426D0995F53F
	//Hash[  1]: C329E527945B8FE75B3C5E8826755747
	//
	//
	//File size: 2*PARTSIZE
	//File hash: A54C5E562D5E03CA7D77961EB9A745A4
	//Nr. hashs: 3
	//Hash[  0]: B3F5CE2A06BF403BFB9BFFF68BDDC4D9
	//Hash[  1]: 509AA30C9EA8FC136B1159DF2F35B8A9
	//Hash[  2]: 31D6CFE0D16AE931B73C59D7E0C089C0	*special part hash*
	//
	//
	//File size: 3*PARTSIZE
	//File hash: 5E249B96F9A46A18FC2489B005BF2667
	//Nr. hashs: 4
	//Hash[  0]: 5319896A2ECAD43BF17E2E3575278E72
	//Hash[  1]: D86EF157D5E49C5ED502EDC15BB5F82B
	//Hash[  2]: 10F2D5B1FCB95C0840519C58D708480F
	//Hash[  3]: 31D6CFE0D16AE931B73C59D7E0C089C0	*special part hash*
	//
	//
	//File size: 3*PARTSIZE + 1 byte
	//File hash: 797ED552F34380CAFF8C958207E40355
	//Nr. hashs: 4
	//Hash[  0]: FC7FD02CCD6987DCF1421F4C0AF94FB8
	//Hash[  1]: 2FE466AF8A7C06DA3365317B75A5ACFE
	//Hash[  2]: 873D3BF52629F7C1527C6E8E473C1C30
	//Hash[  3]: BCE50BEE7877BB07BB6FDA56BFE142FB
	//

	//   File size       Data parts      ED2K parts      ED2K part hashs
	// ---------------------------------------------------------------
	// 1..PARTSIZE-1     1               1                       0(!)
	// PARTSIZE            1               2(!)                   2(!)
	// PARTSIZE+1        2               2                       2
	// PARTSIZE*2        2               3(!)                   3(!)
	// PARTSIZE*2+1    3               3                       3

	if (dwFileSize == 0)
	{
		m_uPartCount = 0;
		m_dwED2KPartCount = 0;
		m_dwED2KPartHashCount = 0;
		m_dwBlockCount = 0;
		m_dwLastPartSz = 0;
		m_dwLastBlkSz = 0;
		return;
	}

	m_dwLastPartSz = dwFileSize % PARTSIZE;
	m_uPartCount = (dwFileSize / PARTSIZE) + ((m_dwLastPartSz != 0) ? 1 : 0);
	m_dwLastPartSz = (m_dwLastPartSz != 0) ? m_dwLastPartSz : PARTSIZE;

	m_dwED2KPartCount = dwFileSize / PARTSIZE + 1;

	m_dwED2KPartHashCount = dwFileSize / PARTSIZE;
	if (m_dwED2KPartHashCount != 0)
	{
		m_dwED2KPartHashCount++;
	}

//	Block traffic
	m_dwLastBlkSz = dwFileSize % EMBLOCKSIZE;
	m_dwBlockCount = (dwFileSize / EMBLOCKSIZE) + ((m_dwLastBlkSz != 0) ? 1 : 0);
	m_dwLastBlkSz = (m_dwLastBlkSz != 0) ? m_dwLastBlkSz : EMBLOCKSIZE;

//	Calculate size ration (=SmallFilePushRatio)
	if (dwFileSize > 0 && dwFileSize < 9961472)
	{
		m_dblSizeRatio = 9961472.0 / static_cast<double>(dwFileSize);
		if (m_dblSizeRatio > 100.0)
		{
			m_dblSizeRatio = 100.0;
		}
	}

//	Intialize array with required status: all parts shared
//	Size must be m_dwED2KPartCount as WritePartStatus sends also status of "void" chunk when filesize % PARTSIZE = 0
	m_PartsStatusVector.resize(m_dwED2KPartCount, 0x00);

}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::SetFileName(const CString& NewName, bool bClearName)
{
//	Set the name & extension
	CAbstractFile::SetFileName(NewName, bClearName);
//	Update full file name
	m_strFilePath.Format(_T("%s\\%s"), m_strKnownFileDirectory, m_strFileName);	//SetFilePath()

//	Don't try to compress already compressed files
	m_bIsCompressedTransferAllowed = ((GetFileType() != ED2KFT_ARCHIVE) || (m_strFileExtension == _T("tar")));
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CKnownFile::CreateFromFile(const CString &strDirectory,const CString &strFileName)
{
	EMULE_TRY

//	TODO: Errorhandling

//
//	First create the filehashset
//

//
//	Try to open the specified file
	CFile					file;
	CArray<uchar*,uchar*>	partHashArray;
	uchar					fileHash[16];
	CString					strNameBuffer;

	SetPath(strDirectory);
	strNameBuffer.Format(_T("%s\\%s"), strDirectory, strFileName);
	SetFilePath(strNameBuffer);

//	If we failed to open the specified file, return false
	if ( !file.Open(strNameBuffer, CFile::modeRead | CFile::shareDenyNone |
		CFile::typeBinary | CFile::osSequentialScan) )
		return false;

	ULONGLONG qwFileSz = file.GetLength();

	if (qwFileSz >= (ULONGLONG)0xFFFFFFFF)
	{
		file.Close();
		return false;
	}

	AddDebugLogLine(RGB_LOG_DIMMED + _T("File %s is being hashed"), strFileName/*m_strFileName*/);

	//set member variables
	SetFileSize(static_cast<uint32>(qwFileSz));
	SetFileName(strFileName, false);
//	after filesize(m_lFileSize) was defined we can get number of hashs
	const uint32 dwHashCount = GetED2KPartHashCount();

//	Create hashset
	if (dwHashCount > 1)
	{
		//	allocate memory for hash array
		partHashArray.SetSize(dwHashCount);
		//	allocate memory for hashset
		uchar*	pPartsHashSet = new uchar[16*dwHashCount];

		for (uint32 i = 0; i < dwHashCount; i++)
		{
		//	get hash over global pointer
			uchar* pNewPartHash = pPartsHashSet + 16*i;

		//	Create a hash for the next part
			if (i == dwHashCount -1)
			{
				CreateHashFromFile(&file, (GetFileSize() - PARTSIZE*i), pNewPartHash);
			}
			else
			{
				CreateHashFromFile(&file, PARTSIZE, pNewPartHash);
			}
#ifndef NEW_SOCKETS_ENGINE
			if (!g_eMuleApp.m_pdlgEmule->IsRunning())	// in case of shutdown while still hashing
			{
				file.Close();
				delete pPartsHashSet;
				return false;
			}
#endif //NEW_SOCKETS_ENGINE
			partHashArray.SetAt(i, pNewPartHash);
		}
		CreateHashFromString(pPartsHashSet, dwHashCount*16, fileHash);

		//copy(add) data from local to global
		CSingleLock Lock(&m_mutexHashList,TRUE);
		m_pPartsHashSet = pPartsHashSet;
		m_partHashArray.SetSize(dwHashCount);
		for (uint32 i = 0; i < dwHashCount; i++)
		{
			m_partHashArray.SetAt(i, partHashArray[i]);
		}
		md4cpy(m_fileHash, fileHash);
		Lock.Unlock();
	}
	else if (dwHashCount == 0)
	{
		//eklmn: hash creation was splitted to 2 steps, cause fuction CreateHashFromFile()
		// needs some time to finish a hashing process . Thefore to prevent possible
		// interference we create a local instance of file hash & then copy it into global
		CreateHashFromFile(&file, GetFileSize(), fileHash);
		md4cpy(m_fileHash, fileHash);
	}
	// TODO: Add filetags

	file.Close();

	struct _stat st;

//	Set last write time
	if (_tstat(strNameBuffer, &st) == 0)
	{
		m_timetLastWriteDate = st.st_mtime;
		AdjustNTFSDaylightFileTime(&m_timetLastWriteDate, strNameBuffer);
	}

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	LoadHashsetFromFile() loads the KnownFile with a hashset from the file 'file' in the format of "known.met".
//		If 'bCheckHash' is true, the file hash in 'file' must match a file hash calculated from the part hashes
//		'file'. false is returned on failure, true on success.
bool CKnownFile::LoadHashsetFromFile(CFile &file, bool bCheckHash)
{
//	Needed for memfiles. It's probably better to switch everything to CFile...
	EMULE_TRY

	uchar					FileHash[16];
	uchar* 					pPartsHashSet = NULL;
	CArray<uchar*,uchar*>	partHashArray;
	uint16					uNumParts;

//	Read the file hash if required, or just skip it if we need just check a hash
	if (!bCheckHash)
		file.Read(&m_fileHash,16);			// <HASH> file hash
	else
		file.Seek(16, CFile::current);

//	Read the number of parts
	file.Read(&uNumParts,2);		// <count:WORD> number of parts

	INT_PTR		iParts = static_cast<uint16>(uNumParts);

	if (iParts < 0)		//Cax2 - why? MOREVIT - Yes, really. Why?? Not used elsewhere.
	{
		if (!m_strFileName.IsEmpty() )
		{
		//	TODO: Add some way to tell if the file is really "known.met" or a memfile.
			AddDebugLogLine(_T("Loaded hashset size is corrupt in known.met, file: %s!"), m_strFileName);
		}
		else
		{
			AddDebugLogLine(_T("Loaded hashset size is corrupt in known.met!"));
		}
	}

	//load hash array in to local buffer
	if (uNumParts == 0)
	{
		if (!bCheckHash)
			return true;	// load was successfully
		else
			return false;	// any check for hash set is impossible
	}
	else		// need to load hashset
	{
	//	If the number of hashes doesn't match the number of hashes the file should have...
		if (bCheckHash && static_cast<uint32>(uNumParts) != GetED2KPartHashCount())
		{
		//	move a pointer to the begin of the file tags
			file.Seek(uNumParts*16, CFile::current);
			return false;
		}

	//	allocate memory for hash array
		partHashArray.SetSize(uNumParts);
	//	allocate memory for hashset
		pPartsHashSet = new uchar[16*uNumParts];

	//	Read the part hashes.
		for (int i = 0; i < static_cast<int>(uNumParts); i++)
		{
			//	get hash over global pointer
			uchar*	pPartHash = pPartsHashSet + 16*i;

			file.Read(pPartHash,16);		// <HASH>[count] part hashes
			partHashArray.SetAt(i, pPartHash);
		}

	//	recreate a file hash from hash set
		CreateHashFromString(pPartsHashSet, uNumParts*16, FileHash);

	//	If the file hash from 'file' matches the calculated hash...
		if (md4cmp(m_fileHash,FileHash) == 0)
		{
			//lock access to global variables & change them
			CSingleLock Lock(&m_mutexHashList,TRUE);
			//check if hash set was allready loaded & delete previous one
			safe_delete(m_pPartsHashSet);
			m_pPartsHashSet = pPartsHashSet;
			m_partHashArray.SetSize(uNumParts);
			for (uint32 i = 0; i < static_cast<int>(uNumParts); i++)
			{
				m_partHashArray.SetAt(i, partHashArray[i]);
			}
			Lock.Unlock();
			return true;
		}
		else
		{
			safe_delete(pPartsHashSet);
			partHashArray.RemoveAll();
			return false;
		}
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	LoadTagsFromFile() parses what tags we know and stores the rest in 'm_tagArray' so they can be rewritten.
bool CKnownFile::LoadTagsFromFile(CFile &file)
{
	EMULE_TRY

	uint32		dwNumTags;

	file.Read(&dwNumTags, 4);							// {<TAGCOUNT:DWORD>

	for (uint32 j = 0; j < dwNumTags; j++)
	{
		CTag		*newtag = new CTag();

		newtag->FillFromStream(file);
		switch (newtag->GetTagID())
		{
			case FT_FILENAME:							// (FT_FILENAME:string) file name
				if (newtag->IsStr())
				{
#ifdef _UNICODE
					if (IsFileNameEmpty())
#endif
						SetFileName(newtag->GetStringValue());
				}
				break;

			case FT_FILESIZE:							// (FT_FILESIZE:int) file size (bytes)
				if (newtag->IsInt())
					SetFileSize(newtag->GetIntValue());
				break;

			case FT_ATTRANSFERRED:						// (FT_ATTRANSFERRED:int) all time transferred (low long)
				if (newtag->IsInt())
					statistic.alltimetransferred = (statistic.alltimetransferred & (0xFFFFFFFFi64 << 32)) | newtag->GetIntValue();
				break;

			case FT_ATTRANSFERREDHI:					// (FT_ATTRANSFERREDHI:int) all time transferred (high long)
				if (newtag->IsInt())
					statistic.alltimetransferred = (statistic.alltimetransferred & 0xFFFFFFFF) |
						(static_cast<uint64>(newtag->GetIntValue()) << 32);
				break;

			case FT_ATREQUESTED:						// (FT_ATREQUESTED:int) all time requested
				if (newtag->IsInt())
					statistic.alltimerequested = newtag->GetIntValue();
				break;

			case FT_ATACCEPTED:							// (FT_ATACCEPTED:int) all time accepted
				if (newtag->IsInt())
					statistic.alltimeaccepted = newtag->GetIntValue();
				break;

#if 1
			case 0x13:	//previously used value left in v1.1c for smooth upgrade, must be removed in v1.1d
#endif
			case FT_ULPRIORITY:							// (FT_ULPRIORITY:int) upload priority
				if (newtag->IsInt())
				{
					m_bytePriority = newtag->GetIntValue();
					if (m_bytePriority == PR_AUTO)
					{
						SetAutoULPriority(true);
						SetULPriority(PR_RELEASE);
					}
					else
					{
						SetAutoULPriority(false);
						if ( (m_bytePriority != PR_VERYLOW) && (m_bytePriority != PR_LOW) &&
							(m_bytePriority != PR_NORMAL) && (m_bytePriority != PR_HIGH) && (m_bytePriority != PR_RELEASE) )
						{
							m_bytePriority = PR_NORMAL;
						}
					}
				}
				break;

			case FT_PERMISSIONS:						// (FT_PERMISSIONS:int) upload permissions
				if (newtag->IsInt() && (newtag->GetIntValue() <= PERM_NOONE))
					m_bytePermissions = newtag->GetIntValue();
				break;

			case FT_PARTFILENAME:	//	This tag is used to keep part file name, delete it as file's already complete
				break;

			default:
				m_tagArray.Add(newtag);
				continue;
		}
		delete newtag;
	}
	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CKnownFile::LoadDateFromFile(CFile& file)
{
	EMULE_TRY

	file.Read(&m_timetLastWriteDate,4);
	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CKnownFile::LoadFromFile(CFile& file)
{
	EMULE_TRY

	bool		result, hashset_ok;

	result = LoadDateFromFile(file);				// <DATE:time_t> last write date/time
	hashset_ok = LoadHashsetFromFile(file,false);	// (HASHSET)
	result = result && LoadTagsFromFile(file);		// <Tag_set>
	result = result && hashset_ok && (GetED2KPartHashCount() == m_partHashArray.GetCount());

//	Cax2 - if m_strFileName==NULL, don't report it, it should still be ok...
#ifndef NEW_SOCKETS_ENGINE
	if(!hashset_ok && !m_strFileName.IsEmpty())
		AddLogLine(false, RGB_LOG_WARNING + GetResString(IDS_ERR_KNOWNHASHCORR), m_strFileName);
#endif //NEW_SOCKETS_ENGINE

#ifndef NEW_SOCKETS_ENGINE
	if (result && CJumpstarter::ShouldBeEnabledForFile(this))
	{
		m_Jumpstarter = new CJumpstarter(this);
		AddLogLine(false, IDS_JS_ENABLED, m_strFileName);
	}
#endif //NEW_SOCKETS_ENGINE

	return result;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CKnownFile::WriteToFile(CFile &file)
{
	byte		abyteBuf[22];
	CSingleLock	Lock(&m_mutexHashList,TRUE);
	CWrTag		tagWr;

	uint32	dwNumParts = m_partHashArray.GetCount();

	if ((dwNumParts != 0) && (m_pPartsHashSet == NULL))
		dwNumParts = 0;

	POKE_DWORD(&abyteBuf[0], m_timetLastWriteDate);				// <DATE:time_t> last write date/time
	md4cpy(&abyteBuf[4], m_fileHash);							// <HASH>
	POKE_WORD(&abyteBuf[20], static_cast<uint16>(dwNumParts));	// <count:WORD> number of chunk hashes

	file.Write(abyteBuf, sizeof(abyteBuf));
	if (m_pPartsHashSet != NULL)
		file.Write(m_pPartsHashSet, 16 * dwNumParts);			// <HASH>[count] part hashes

	Lock.Unlock();

//	Tags
	uint32	dwTagFilePos = static_cast<uint32>(file.GetPosition());
	uint32	dwTmp, dwTagCnt = 0;

	file.Write(&dwTagCnt, 4);									// <TAGCOUNT:DWORD>

	tagWr.WriteToFile(FT_FILENAME, GetFileName(), file);		// (FT_FILENAME:string) file name
	dwTagCnt++;
	tagWr.WriteToFile(FT_FILESIZE, GetFileSize(), file);		// (FT_FILESIZE:int) file size (bytes)
	dwTagCnt++;
	if (statistic.GetAllTimeTransferred() != 0)
	{
		tagWr.WriteToFile(FT_ATTRANSFERRED, static_cast<uint32>(statistic.GetAllTimeTransferred()), file);
		if ((dwTmp = static_cast<uint32>(statistic.GetAllTimeTransferred() >> 32)) != 0)
		{
			tagWr.WriteToFile(FT_ATTRANSFERREDHI, dwTmp, file);
			dwTagCnt++;
		}
		dwTagCnt++;
	}
	if (statistic.GetAllTimeRequests() != 0)
	{
		tagWr.WriteToFile(FT_ATREQUESTED, statistic.GetAllTimeRequests(), file);
		dwTagCnt++;
	}
	if (statistic.GetAllTimeAccepts() != 0)
	{
		tagWr.WriteToFile(FT_ATACCEPTED, statistic.GetAllTimeAccepts(), file);
		dwTagCnt++;
	}
	tagWr.WriteToFile(FT_ULPRIORITY, IsULAutoPrioritized() ? PR_AUTO : m_bytePriority, file);
	dwTagCnt++;
	if (m_bytePermissions != PERM_ALL)
	{
		tagWr.WriteToFile(FT_PERMISSIONS, m_bytePermissions, file);
		dwTagCnt++;
	}

//	Unknown tags
	for (int j = 0; j < m_tagArray.GetCount(); j++)
	{
		if (m_tagArray[j]->IsStr() || m_tagArray[j]->IsInt())
		{
			m_tagArray[j]->WriteToFile(file);
			dwTagCnt++;
		}
	}
//	Save valid tag count
	file.Seek(dwTagFilePos, CFile::begin);
	file.Write(&dwTagCnt, 4);
	file.SeekToEnd();

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	CreateHashFromInput() generates a hash from the next 'dwLength' bytes of 'file' or 'pbyteMem'
//		The hash is returned in '*pbyteHash'.
void CKnownFile::CreateHashFromInput(CFile *file, uint32 dwLength, uchar *pbyteHash, uchar *pbyteMem)
{
	EMULE_TRY

	uint32 *pdwHash = reinterpret_cast<uint32*>(pbyteHash);

	pdwHash[0] = 0x67452301;
	pdwHash[1] = 0xEFCDAB89;
	pdwHash[2] = 0x98BADCFE;
	pdwHash[3] = 0x10325476;

	uint32		dwRequired = dwLength;
	uchar		pbyteBuff[1024];	//	should be small enough to eliminate stack size check

	if (pbyteMem != NULL)
	{
	//	Process memory block
		while (dwRequired >= 64)
		{
			MD4Transform(pdwHash, reinterpret_cast<uint32*>(pbyteMem));
			pbyteMem += 64;
			dwRequired -= 64;
		}
	//	Copy the rest
		if (dwRequired != 0)
			memcpy2(pbyteBuff, pbyteMem, dwRequired);
	}
	else
	{
	//	Process a file
		uint32		dwBlockSz = 16 * 4096;
		void		*pAlloc = malloc(dwBlockSz);
		uchar		*pbytePtr = reinterpret_cast<uchar*>(pAlloc);

		if (pbytePtr == NULL)
		{
		//	Use small local buffer in case of low memory
			pbytePtr = pbyteBuff;
			dwBlockSz = sizeof(pbyteBuff);
		}

		while (dwRequired >= dwBlockSz)
		{
			file->Read(pbytePtr, dwBlockSz);
			uint32 i = 0;
			do
			{
				MD4Transform(pdwHash, (uint32*)(pbytePtr + i));
				i += 64;
			} while(i < dwBlockSz);
			dwRequired -= dwBlockSz;
		}
		if (dwRequired >= 64)
		{
			uint32 dwLen = dwRequired & ~63;

			file->Read(pbytePtr, dwLen);
			uint32 i = 0;
			do
			{
				MD4Transform(pdwHash, reinterpret_cast<uint32*>(pbytePtr + i));
				i += 64;
			} while(i < dwLen);
			dwRequired -= dwLen;
		}
	//	Read the rest
		if (dwRequired != 0)
			file->Read(pbyteBuff, dwRequired);
		if (pAlloc != NULL)
			free(pAlloc);
	}
	// in byte scale 512 = 64, 448 = 56
	pbyteBuff[dwRequired++] = 0x80;
	if (dwRequired > 56)
	{
		memzero(&pbyteBuff[dwRequired], 64 - dwRequired);
		MD4Transform(pdwHash, reinterpret_cast<uint32*>(pbyteBuff));
		dwRequired = 0;
	}
	memzero(&pbyteBuff[dwRequired], 56 - dwRequired);
// Add size (convert to bits)
	*reinterpret_cast<uint32*>(&pbyteBuff[56]) = (dwLength << 3);
	*reinterpret_cast<uint32*>(&pbyteBuff[60]) = (dwLength >> 29);

	MD4Transform(pdwHash, reinterpret_cast<uint32*>(pbyteBuff));

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uchar* CKnownFile::GetPartHash(uint16 part)
{
	EMULE_TRY
	if (part >= m_partHashArray.GetCount())
		return 0;

	return m_partHashArray[part];

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double CKnownFile::GetPopularityRatio()
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	if (g_eMuleApp.m_pKnownFilesList->GetTotalRequested()==0)
		return 1.0;
	if (statistic.GetRequests() == 0)
		return 2.0;

	double		dblRatio;

	dblRatio = (static_cast<double>(statistic.GetRequests())) / (static_cast<double>(g_eMuleApp.m_pKnownFilesList->GetTotalRequested()));
	dblRatio = -4.0 * log10(static_cast<double>(g_eMuleApp.m_pSharedFilesList->GetCount())) * dblRatio;
	dblRatio = 2.0 * exp(dblRatio);

	if (dblRatio < 1.0)
	{
		dblRatio = 1.0;
	}
	else if (dblRatio > 2.0)
	{
		dblRatio = 2.0;
	}

	return dblRatio;
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH

	return 1.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
static void __fastcall MD4Transform(uint32 Hash[4], uint32 x[16])
{
	uint32 a = Hash[0];
	uint32 b = Hash[1];
	uint32 c = Hash[2];
	uint32 d = Hash[3];

	/* Round 1 */
	MD4_FF(a, b, c, d, x[ 0], S11); // 01
	MD4_FF(d, a, b, c, x[ 1], S12); // 02
	MD4_FF(c, d, a, b, x[ 2], S13); // 03
	MD4_FF(b, c, d, a, x[ 3], S14); // 04
	MD4_FF(a, b, c, d, x[ 4], S11); // 05
	MD4_FF(d, a, b, c, x[ 5], S12); // 06
	MD4_FF(c, d, a, b, x[ 6], S13); // 07
	MD4_FF(b, c, d, a, x[ 7], S14); // 08
	MD4_FF(a, b, c, d, x[ 8], S11); // 09
	MD4_FF(d, a, b, c, x[ 9], S12); // 10
	MD4_FF(c, d, a, b, x[10], S13); // 11
	MD4_FF(b, c, d, a, x[11], S14); // 12
	MD4_FF(a, b, c, d, x[12], S11); // 13
	MD4_FF(d, a, b, c, x[13], S12); // 14
	MD4_FF(c, d, a, b, x[14], S13); // 15
	MD4_FF(b, c, d, a, x[15], S14); // 16

	/* Round 2 */
	MD4_GG(a, b, c, d, x[ 0], S21); // 17
	MD4_GG(d, a, b, c, x[ 4], S22); // 18
	MD4_GG(c, d, a, b, x[ 8], S23); // 19
	MD4_GG(b, c, d, a, x[12], S24); // 20
	MD4_GG(a, b, c, d, x[ 1], S21); // 21
	MD4_GG(d, a, b, c, x[ 5], S22); // 22
	MD4_GG(c, d, a, b, x[ 9], S23); // 23
	MD4_GG(b, c, d, a, x[13], S24); // 24
	MD4_GG(a, b, c, d, x[ 2], S21); // 25
	MD4_GG(d, a, b, c, x[ 6], S22); // 26
	MD4_GG(c, d, a, b, x[10], S23); // 27
	MD4_GG(b, c, d, a, x[14], S24); // 28
	MD4_GG(a, b, c, d, x[ 3], S21); // 29
	MD4_GG(d, a, b, c, x[ 7], S22); // 30
	MD4_GG(c, d, a, b, x[11], S23); // 31
	MD4_GG(b, c, d, a, x[15], S24); // 32

	/* Round 3 */
	MD4_HH(a, b, c, d, x[ 0], S31); // 33
	MD4_HH(d, a, b, c, x[ 8], S32); // 34
	MD4_HH(c, d, a, b, x[ 4], S33); // 35
	MD4_HH(b, c, d, a, x[12], S34); // 36
	MD4_HH(a, b, c, d, x[ 2], S31); // 37
	MD4_HH(d, a, b, c, x[10], S32); // 38
	MD4_HH(c, d, a, b, x[ 6], S33); // 39
	MD4_HH(b, c, d, a, x[14], S34); // 40
	MD4_HH(a, b, c, d, x[ 1], S31); // 41
	MD4_HH(d, a, b, c, x[ 9], S32); // 42
	MD4_HH(c, d, a, b, x[ 5], S33); // 43
	MD4_HH(b, c, d, a, x[13], S34); // 44
	MD4_HH(a, b, c, d, x[ 3], S31); // 45
	MD4_HH(d, a, b, c, x[11], S32); // 46
	MD4_HH(c, d, a, b, x[ 7], S33); // 47
	MD4_HH(b, c, d, a, x[15], S34); // 48

	Hash[0] += a;
	Hash[1] += b;
	Hash[2] += c;
	Hash[3] += d;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Updates priority of file if autopriority is activated
void CKnownFile::UpdateUploadAutoPriority(void)
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	if (IsULAutoPrioritized() && !IsCompleting())
	{
		if (IsPartFile() && g_eMuleApp.m_pDownloadQueue && !IsPaused() )
		{
			CPartFile		*pPartFile = dynamic_cast<CPartFile*>(this);

			if(pPartFile != NULL)
			{
				int iValidSources = pPartFile->GetSourceCount() - pPartFile->GetNotCurrentSourcesCount();

				if (iValidSources < RARE_FILE)
				{
					SetULPriority(PR_RELEASE);
				}
				else if (iValidSources < 200)
				{
					SetULPriority(PR_HIGH);
				}
				else if (iValidSources < 400)
				{
					SetULPriority(PR_NORMAL);
				}
				else
				{
					SetULPriority(PR_LOW);
				}

			}
		}
		else
		{
			int iCompleteSources = GetCompleteSourcesCount();

			if (iCompleteSources < 5)
			{
				SetULPriority(PR_RELEASE);
			}
			else if (iCompleteSources < RARE_FILE)
			{
				SetULPriority(PR_HIGH);
			}
			else if (iCompleteSources < 200)
			{
				SetULPriority(PR_NORMAL);
			}
			else
			{
				SetULPriority(PR_LOW);
			}
		}
	}
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CAbstractFile::SetFileName(const CString& NewName, bool bClearName)
{
	m_strFileName = NewName;

	if (!m_strFileName.IsEmpty())
	{
		m_eFileType = GetED2KFileTypeID(m_strFileName);

		int		iIdxExtension = m_strFileName.ReverseFind(_T('.'));

		if (iIdxExtension != -1)
		{
			m_strFileExtension = m_strFileName.Mid(iIdxExtension + 1);
			m_strFileExtension.MakeLower();
		}
		else
			m_strFileExtension.Empty();
	}
	else
	{
		m_eFileType = ED2KFT_ANY;
		m_strFileExtension.Empty();
	}

	if (bClearName)
	{
		static const TCHAR s_acRepTbl[] =
		{
			_T('/'), _T('>'), _T('<'), _T('*'), _T(':'), _T('?'), _T('|'), _T('\"'), _T('\\')
		};

		for (unsigned ui = 0; ui < _countof(s_acRepTbl); ui++)
			m_strFileName.Replace(s_acRepTbl[ui], _T('-'));
	}
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CAbstractFile::GetFileTypeString(bool bLocalized)
{
#ifndef NEW_SOCKETS_ENGINE
	static const uint16 s_auResID[ED2KFT_COUNT] =
	{
		IDS_SEARCH_ANY,		// ED2KFT_ANY
		IDS_SEARCH_AUDIO,	// ED2KFT_AUDIO
		IDS_SEARCH_VIDEO,	// ED2KFT_VIDEO
		IDS_SEARCH_PICS,	// ED2KFT_IMAGE
		IDS_SEARCH_PRG,		// ED2KFT_PROGRAM
		IDS_SEARCH_DOC,		// ED2KFT_DOCUMENT
		IDS_SEARCH_ARC,		// ED2KFT_ARCHIVE
		IDS_SEARCH_CDIMG	// ED2KFT_CDIMAGE
	};
	static const char *const s_apcSearchTypes[ED2KFT_COUNT] =
	{
		"",			// ED2KFT_ANY
		"Audio",	// ED2KFT_AUDIO
		"Video",	// ED2KFT_VIDEO
		"Image",	// ED2KFT_IMAGE
		"Pro",		// ED2KFT_PROGRAM
		"Doc",		// ED2KFT_DOCUMENT
		"Pro",		// ED2KFT_ARCHIVE
		"Pro"		// ED2KFT_CDIMAGE
	};
	unsigned	uiType = m_eFileType;

	if (uiType >= ED2KFT_COUNT)
		uiType = ED2KFT_ANY;

	if (bLocalized)
		return GetResString(s_auResID[uiType]);
	else
		return s_apcSearchTypes[uiType];
#else
	return _T("");
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
Packet*	CKnownFile::CreateSrcInfoPacket(CUpDownClient *pForClient)
{
#ifndef NEW_SOCKETS_ENGINE
	EMULE_TRY

	if (g_eMuleApp.m_pUploadQueue->GetWaitingUserCount() == 0)
		return NULL;

	CSafeMemFile	packetStream(1024);
	uint16			uCount = 0;

	packetStream.Write(pForClient->m_reqFileHash, 16);
	packetStream.Write(&uCount, 2);

	CUpDownClient  *pPotentialSource;
	bool 			bNeededPart;
	byte		   *pbyteForClientPartStatus;
	byte		   *pbytePotentialClientPartStatus;

	uint32			dwID;
	uint16			uPort;
	uint32			dwServerIP;
	uint16			uServerPort;

	for (POSITION pos = g_eMuleApp.m_pUploadQueue->GetHeadPosition(); pos != NULL; )
	{
		pPotentialSource = g_eMuleApp.m_pUploadQueue->GetNext(pos);

	//	Don't send a client to himself as a source
	//	Only send sources with LowID that are in same sever that client who is requesting sources
	//	Don't send clients that have no parts available
	//	Don't share Lan clients sources as they are private ips/userids
		if ( md4cmp(pPotentialSource->m_reqFileHash, pForClient->m_reqFileHash) != 0
				|| pForClient == pPotentialSource || pForClient->Compare(pPotentialSource)
				|| ( pPotentialSource->HasLowID()
					&& ( pPotentialSource->GetServerIP() != pForClient->GetServerIP()
						|| pPotentialSource->GetServerPort() != pForClient->GetServerPort() ) )
				|| (pPotentialSource->GetAvailUpPartCount() == 0 && pForClient->GetUpPartCount() > 1)
				|| pPotentialSource->IsOnLAN()
				|| pForClient->GetUpPartCount() != pPotentialSource->GetUpPartCount() )
			continue;

		bNeededPart = false;

	//	Only send source which have needed parts for this client if possible
		pbyteForClientPartStatus = pForClient->GetPartStatus();
		pbytePotentialClientPartStatus = pPotentialSource->GetPartStatus();

	//	Both clients support part statuses so we try to find needed clients
		if (pbyteForClientPartStatus != NULL && pbytePotentialClientPartStatus != NULL)
		{
			for (uint32 i = 0; i < pForClient->GetUpPartCount(); i++)
			{
				if(pbytePotentialClientPartStatus[i] && !pbyteForClientPartStatus[i])
				{
					bNeededPart = true;
					break;
				}
			}
		}
	//	If not we give a sources with at least one part (for files that have more than one part)
		else
			bNeededPart = true;

		if (bNeededPart)
		{
			uCount++;
			dwID = pPotentialSource->GetUserID();
			uPort = pPotentialSource->GetUserPort();
			dwServerIP = pPotentialSource->GetServerIP();
			uServerPort = pPotentialSource->GetServerPort();

			packetStream.Write(&dwID, 4);
			packetStream.Write(&uPort, 2);
			packetStream.Write(&dwServerIP, 4);
			packetStream.Write(&uServerPort, 2);
			if (pForClient->GetSourceExchangeVersion() > 1)
			    packetStream.Write(pPotentialSource->GetUserHash(),16);
			if (uCount > 500)
				break;
		}
	}

	if (uCount == 0)
		return NULL;

	packetStream.Seek(16, 0);
	packetStream.Write(&uCount, 2);

	Packet	*pPacket = new Packet(&packetStream,OP_EMULEPROT);

	pPacket->m_eOpcode = OP_ANSWERSOURCES;
	if (pPacket->m_dwSize > 354)
		pPacket->PackPacket();

	return pPacket;

	EMULE_CATCH
#endif //NEW_SOCKETS_ENGINE

	return NULL;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	For File Comment
void CKnownFile::LoadComment()
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	CString	strFullPath;

	strFullPath.Format(_T("%sfileinfo.ini"),g_eMuleApp.m_pGlobPrefs->GetConfigDir());

	CIni ini(strFullPath);

	ini.SetDefaultCategory(HashToString(m_fileHash));
	m_strComment = ini.GetString(_T("Comment"),_T(""));
	m_eRating = static_cast<_EnumPartFileRating>(ini.GetInt(_T("Rate"), PF_RATING_NONE));
	m_bCommentLoaded = true;

	ini.CloseWithoutSave();
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::RemoveFileCommentAndRating()
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	CString strFullPath;

	strFullPath.Format(_T("%sfileinfo.ini"),g_eMuleApp.m_pGlobPrefs->GetConfigDir());

	CIni ini(strFullPath);

	ini.DeleteCategory(HashToString(m_fileHash));
	ini.SaveAndClose();

	m_strComment = _T("");
	m_eRating = PF_RATING_NONE;

	if (IsPartFile())
	{
		((CPartFile*)this)->SetHasComment(false);
		((CPartFile*)this)->SetHasRating(false);
	}
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::SetFileComment(CString strNewComment)
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	CString	strFullPath;

	strFullPath.Format(_T("%sfileinfo.ini"),g_eMuleApp.m_pGlobPrefs->GetConfigDir());

	CIni ini(strFullPath);

	ini.SetDefaultCategory(HashToString(m_fileHash));

	ini.SetString (_T("Comment"), strNewComment);
	m_strComment = strNewComment;

	ini.SaveAndClose();

	CTypedPtrList<CPtrList, CUpDownClient*> srclist;

	g_eMuleApp.m_pUploadQueue->FindSourcesForFileById(&srclist, GetFileHash());

	CUpDownClient	*pClient;

	for (POSITION pos = srclist.GetHeadPosition(); pos != NULL; )
	{
		pClient = srclist.GetNext(pos);
		pClient->SetCommentDirty();
	}

	if (IsPartFile())
		((CPartFile*)this)->SetHasComment(true);
#endif //NEW_SOCKETS_ENGINE

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	For File rate
void CKnownFile::SetFileRating(EnumPartFileRating eNewRating)
{
	EMULE_TRY

#ifndef NEW_SOCKETS_ENGINE
	CString strFullPath;

	strFullPath.Format(_T("%sfileinfo.ini"),g_eMuleApp.m_pGlobPrefs->GetConfigDir());

	CIni ini(strFullPath);

	ini.SetDefaultCategory(HashToString(m_fileHash));

	ini.SetInt (_T("Rate"), eNewRating);
	m_eRating = eNewRating;

	ini.SaveAndClose();

	CTypedPtrList<CPtrList, CUpDownClient*> srclist;
	g_eMuleApp.m_pUploadQueue->FindSourcesForFileById(&srclist, GetFileHash());

	CUpDownClient	*pClient;

	for (POSITION pos = srclist.GetHeadPosition(); pos != NULL; )
	{
		pClient = srclist.GetNext(pos);
		pClient->SetCommentDirty();
	}

	if (IsPartFile())
		((CPartFile*)this)->SetHasRating(true);

#endif //NEW_SOCKETS_ENGINE
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::initTraffic()
{
	EMULE_TRY

	if(partTraffic==NULL)
	{
		uint16 p=fileParent->GetPartCount();
		partTraffic=new uint32[p];
		partTrafficSession=new uint32[p];
		for(uint16 i=0; i<p; i++)
		{
			partTraffic[i]=0;
			partTrafficSession[i]=0;
		}
	}

	if(partAccepted==NULL)
	{
		uint16 p=fileParent->GetPartCount();
		partAccepted=new uint32[p];
		partAcceptedSession=new uint32[p];
		for(uint16 i=0; i<p; i++)
		{
			partAccepted[i]=0;
			partAcceptedSession[i]=0;
		}
	}

	if(blockTraffic==NULL)
	{
		uint32 p=fileParent->GetBlockCount();
		blockTraffic=new uint32[p];
		blockTrafficSession=new uint32[p];
		for(uint32 i=0; i<p; i++)
		{
			blockTraffic[i]=0;
			blockTrafficSession[i]=0;
		}
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::recalcCompleteReleases()
{
	EMULE_TRY

//	Recalculate complete releases factor
	uint32 blocks=fileParent->GetBlockCount();
	uint32 fs=fileParent->GetFileSize();

//	Find lowest parttraffic
	uint32 m=0xFFFFFFFF;
	completeReleases=0;
	for(uint32 i=0; i<blocks; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s = ((fs - (i * EMBLOCKSIZE)) >= EMBLOCKSIZE) ? EMBLOCKSIZE : (fs % EMBLOCKSIZE);

		if(s)
		{
			uint32 f=(uint32)(double)blockTraffic[i]/s+(blockTraffic[i]==s?0:1);
			if(f<m) m=f;
		}
	}

//	Get average
	for(uint32 i=0; i<blocks; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s = ((fs - (i * EMBLOCKSIZE)) >= EMBLOCKSIZE) ? EMBLOCKSIZE : (fs % EMBLOCKSIZE);

		if(s)
		{
			double f=(double)blockTraffic[i]/s;
			if(f>m) f=m;
			completeReleases+=f;
		}
	}

	completeReleases/=blocks;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	This function may not return a 100% accurate value, because you can't fill a part with blocks
double CFileStatistic::GetCompletePartReleases(uint16 part)
{
	EMULE_TRY

//	Check part
	if(partTraffic==NULL || part<0 || part>fileParent->GetPartCount() || partTraffic[part]==0)
		return 0;

//	Start and end block
	uint32	fs = fileParent->GetFileSize();
	uint32	startBlock=(part*PARTSIZE)/EMBLOCKSIZE;
	uint32	endBlock = ((fs - (part * PARTSIZE)) > PARTSIZE) ? (((part + 1) * PARTSIZE) / EMBLOCKSIZE) : (fileParent->GetBlockCount() - 1);

//	Find lowest parttraffic
	uint32	m=0xFFFFFFFF;
	double	completePartReleases=0;
	for(uint32 i=startBlock; i<=endBlock; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s = ((fs - (i * EMBLOCKSIZE)) >= EMBLOCKSIZE) ? EMBLOCKSIZE : (fs % EMBLOCKSIZE);

		if(s)
		{
			uint32 f=(uint32)(double)blockTraffic[i]/s+(blockTraffic[i]==s?0:1);
			if(f<m) m=f;
		}
	}

//	Get avg.
	for(uint32 i=startBlock; i<=endBlock; i++)
	{
		if(blockTraffic[i]==0)
			continue;

		uint32 s = ((fs - (i * EMBLOCKSIZE)) >= EMBLOCKSIZE) ? EMBLOCKSIZE : (fs % EMBLOCKSIZE);

		if(s)
		{
			double f=(double)blockTraffic[i]/s;
			if(f>m) f=m;
			completePartReleases+=f;
		}
	}

	return completePartReleases/(fileParent->GetPartSize(part)/EMBLOCKSIZE);

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::AddTraffic(uint32 start, uint32 bytes)
{
	EMULE_TRY

	initTraffic();

//	Start block
	uint32	part=start/PARTSIZE;
	sint32	btg=(sint32)bytes;

	uint32 p1=min((part+1)*PARTSIZE-start, bytes);
	partTraffic[part]+=p1;
	partTrafficSession[part]+=p1;
	partAccepted[part]++;
	partAcceptedSession[part]++;
	btg-=p1;
	part++;
	while(btg>0)
	{
		partTraffic[part]+=btg % PARTSIZE;
		partTrafficSession[part]+=btg % PARTSIZE;
		partAccepted[part]++;
		partAcceptedSession[part]++;
		part++;
		btg-=PARTSIZE;
	}

//	Start block
	uint32	block=start/EMBLOCKSIZE;
	btg=(sint32)bytes;

	uint32 b1=min((block+1)*EMBLOCKSIZE-start, bytes);
	blockTraffic[block]+=b1;
	blockTrafficSession[block]+=b1;
	btg-=b1;
	block++;
	while(btg>0)
	{
		blockTraffic[block]+=btg % EMBLOCKSIZE;
		blockTrafficSession[block]+=btg % EMBLOCKSIZE;
		block++;
		btg-=EMBLOCKSIZE;
	}

	recalcCompleteReleases();

	fileParent->UpdateSharedFileDisplay();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CKnownFile::SaveToFileTraffic(FILE* file)
{
	EMULE_TRY

	if(statistic.partTraffic)
	{
		uint16 parts=GetPartCount();
		if(fwrite(&parts, sizeof(parts), 1, file)!=1) return false;
		if(fwrite(statistic.partTraffic, 4, parts, file)!=parts) return false;

		uint32 blocks=GetBlockCount();
		if(fwrite(&blocks, sizeof(blocks), 1, file)!=1) return false;
		if(fwrite(statistic.blockTraffic, 4, blocks, file)!=blocks) return false;

		if(fwrite(statistic.partAccepted, 4, parts, file)!=parts) return false;
	}

//	This block shouldnt be used at all
	else
	{
		uint32	d=0;

		uint16 parts=GetPartCount();
		if(fwrite(&parts, sizeof(parts), 1, file)!=1) return false;
		for(uint16 part=0; part<parts; part++)
			if(fwrite(&d, sizeof(d), 1, file)!=1) return false;

		uint32 blocks=GetBlockCount();
		if(fwrite(&blocks, sizeof(blocks), 1, file)!=1) return false;
		for(uint32 block=0; block<blocks; block++)
			if(fwrite(&d, sizeof(d), 1, file)!=1) return false;

		for(uint16 part=0; part<parts; part++)
			if(fwrite(&d, sizeof(d), 1, file)!=1) return false;
	}

	return ferror(file);

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CKnownFile::LoadFromFileTraffic(FILE* file, byte version)
{
	EMULE_TRY

	// read parts
	uint16 parts=0;
	if(fread(&parts, sizeof(parts), 1, file)!=1) return false;

	// does part count match?
	if(parts!=GetPartCount())
	{
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_TRAFFICDAT_CORRUPT));
		g_eMuleApp.m_pdlgEmule->DisableAutoBackup();
#endif //NEW_SOCKETS_ENGINE

		return false;
	}

	// get the memory
	statistic.initTraffic();

	// load all the parts traffic data
	if(fread(statistic.partTraffic, 4, parts, file)!=parts) return false;

	// read blocks
	uint32 blocks=0;
	if(fread(&blocks, sizeof(blocks), 1, file)!=1) return false;

	// does blocks count match?
	if(blocks!=GetBlockCount())
	{
#ifndef NEW_SOCKETS_ENGINE
		AddLogLine(true, RGB_LOG_ERROR + GetResString(IDS_TRAFFICDAT_CORRUPT));
		g_eMuleApp.m_pdlgEmule->DisableAutoBackup();
#endif //NEW_SOCKETS_ENGINE

		return false;
	}

	// load all the parts traffic data
	if(fread(statistic.blockTraffic, 4, blocks, file)!=blocks) return false;

//	Versioning
	if(version>1)
	{
	//	Version 2 introduced part-accepted counter
		if(fread(statistic.partAccepted, 4, parts, file)!=parts) return false;
	}

//	Recalc something ---
	statistic.recalcCompleteReleases();

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CKnownFile::GetTrafficBlock(uint32 start, uint32 end, bool session)
{
	EMULE_TRY

	if(statistic.blockTraffic==NULL)
		return 0;

	if(start > GetFileSize()) return 0;
	if(end > GetFileSize()) end=GetFileSize();

	uint32	*data;
	if(session)
		data=statistic.blockTrafficSession;
	else
		data=statistic.blockTraffic;

	uint32	block=start/EMBLOCKSIZE;
	uint32	pos=start;
	sint32	bftb;
	uint32	t=0;
	while(pos < end)
	{
		if(end-pos > end % EMBLOCKSIZE)
			bftb=EMBLOCKSIZE;
		else
			bftb=end % EMBLOCKSIZE;
		bftb-=pos % EMBLOCKSIZE;

		if(bftb != EMBLOCKSIZE)
			t+=data[block] * (double)bftb/GetBlockSize(block);
		else
			t+=data[block];

		pos+=bftb;

		block++;
	}

	return t;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CKnownFile::GetTrafficPart(uint32 start, uint32 end, bool session)
{
	EMULE_TRY

	if(statistic.partTraffic==NULL)
		return 0;

	if(start > GetFileSize()) return 0;
	if(end > GetFileSize()) end=GetFileSize();

	uint32	*data;
	if(session)
		data=statistic.partTrafficSession;
	else
		data=statistic.partTraffic;

	uint32	part=start/PARTSIZE;
	uint32	pos=start;
	sint32	bftp;
	uint32	t=0;
	while(pos < end)
	{
		if(end-pos > end % PARTSIZE)
			bftp=PARTSIZE;
		else
			bftp=end % PARTSIZE;
		bftp-=pos % PARTSIZE;

		if(bftp != PARTSIZE)
			t+=data[part] * (double)bftp/GetPartSize(part);
		else
			t+=data[part];

		pos+=bftp;

		part++;
	}

	return t;

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CKnownFile::GetPartTraffic(uint16 part, bool session)
{
	EMULE_TRY

	if(statistic.partTraffic==NULL)
		return 0;

	if(session)
		return statistic.partTrafficSession[part];
	else
		return statistic.partTraffic[part];

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CKnownFile::GetBlockTraffic(uint16 block, bool session)
{
	EMULE_TRY

	if(statistic.partTraffic==NULL)
		return 0;

	if(session)
		return statistic.blockTrafficSession[block];
	else
		return statistic.blockTraffic[block];

	EMULE_CATCH

	return 0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CKnownFile::GetPartSize(uint32 dwPart) const
{
	return ((dwPart + 1) < GetPartCount()) ? PARTSIZE : m_dwLastPartSz;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CKnownFile::GetBlockSize(uint32 dwBlock) const
{
	return ((dwBlock + 1) < GetBlockCount()) ? EMBLOCKSIZE : m_dwLastBlkSz;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::resetStats(bool all)
{
	EMULE_TRY

	if (all)
	{
		if (partTraffic)		delete[] partTraffic;	partTraffic = NULL;
		if (partAccepted)	delete[] partAccepted;	partAccepted = NULL;
		if (blockTraffic)	delete[] blockTraffic;	blockTraffic = NULL;

		if (partTrafficSession)		delete[] partTrafficSession;	partTrafficSession = NULL;
		if (partAcceptedSession)		delete[] partAcceptedSession;	partAcceptedSession = NULL;
		if (blockTrafficSession)		delete[] blockTrafficSession;	blockTrafficSession = NULL;

		completeReleases = 0.0;

		alltimerequested = alltimetransferred = alltimeaccepted=0;
	}
	else
	{
		if (partTrafficSession)
			for (uint16 part = 0; part < fileParent->GetPartCount(); part++)
				partTrafficSession[part] = 0;

		if (partAcceptedSession)
			for (uint16 part = 0; part < fileParent->GetPartCount(); part++)
				partAcceptedSession[part] = 0;

		if (blockTrafficSession)
			for (uint32 block = 0; block < fileParent->GetBlockCount(); block++)
				blockTrafficSession[block] = 0;
	}

	m_iNumRequested = m_qwNumTransferred=m_iNumAccepted=0;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CFileStatistic::resetPartTraffic(uint16 part, bool all)
{
	EMULE_TRY

	if(partTraffic==NULL || part>=fileParent->GetPartCount())
		return;

	//--- this all is a little bit tricky, because the part and block size have no connections ---
	//--- so i have to remove some bytes only from the start and end block ---
	//--- all we try is to keep it as realistic as possible :) ---
	uint32	blocks=fileParent->GetBlockCount();
	uint32	blockStart=(part*PARTSIZE)/EMBLOCKSIZE;
	uint32	blockEnd=min((((uint32)part+1)*PARTSIZE)/EMBLOCKSIZE+1, blocks-1);
	double	percentStart=(static_cast<double>((part*PARTSIZE)%EMBLOCKSIZE)/EMBLOCKSIZE);
	double	percentEnd=0.0;
	if(blockEnd!=blocks)
		percentEnd=1-(static_cast<double>(((part+1)*PARTSIZE)%EMBLOCKSIZE)/EMBLOCKSIZE);

	if(all)
	{
		partTraffic[part]=0;
		blockTraffic[blockStart]*=percentStart;
		blockTraffic[blockEnd]*=percentEnd;
		for(uint32 block=blockStart+1; block<blockEnd; block++)
			blockTraffic[block]=0;
	}

	partTrafficSession[part]=0;
	blockTrafficSession[blockStart]*=percentStart;
	blockTrafficSession[blockEnd]*=percentEnd;
	for(uint32 block=blockStart+1; block<blockEnd; block++)
		blockTrafficSession[block]=0;

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32	CFileStatistic::GetPartAccepted(uint16 part, bool session) const
{
	EMULE_TRY

	if(partTraffic==NULL || part>=fileParent->GetPartCount())
		return 0;

	if(session==false)
		return partAccepted[part];
	else
		return partAcceptedSession[part];

	EMULE_CATCH

	return 0;
}
//--- :xrmb ---
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//--- xrmb:allyaknow ---
//-------------------------------------------------------------------------------------------------
bool CFileStatistic::merge(CFileStatistic *m)
{
	EMULE_TRY

	alltimerequested	+= m->alltimerequested;
	alltimeaccepted		+= m->alltimeaccepted;
	alltimetransferred	+= m->alltimetransferred;

	m_iNumRequested	+= m->m_iNumRequested;
	m_iNumAccepted	+= m->m_iNumAccepted;
	m_qwNumTransferred	+= m->m_qwNumTransferred;

	//--- merge parttraffic ---

	//--- i dont have traffic, but the to merge has ---
	if(partTraffic==NULL && m->partTraffic)
		initTraffic();

	//--- merge traffic only if the to merge has some ---
	if(m->partTraffic)
	{
		for(uint16 part=0; part<fileParent->GetPartCount(); part++)
			partTraffic[part]+=m->partTraffic[part];

		for(uint32 block=0; block<fileParent->GetBlockCount(); block++)
			blockTraffic[block]+=m->blockTraffic[block];
	}

	return true;

	EMULE_CATCH

	return false;
}
//--- :xrmb ---
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool	CKnownFile::HasHiddenParts()
{
	EMULE_TRY

	for (uint16 i = 0; i < GetPartCount(); i++)
	{
		if (!IsPartShared(i))
			return true;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::WritePartStatus(CFile *file)
{
	EMULE_TRY

	uint16	uParts = m_partHashArray.GetCount();
	uchar	byteTmpBuff[32];	// should be power of 2
	uint32	dwDone = 0, dwBuffIdx = 2, dwDiff, dwParts = static_cast<uint32>(uParts);

	*reinterpret_cast<uint16*>(&byteTmpBuff[0]) = uParts;

	while ((dwDiff = (dwParts - dwDone)) != 0)
	{
		byte	byteMask = 1;
		uint32	dwLimit = dwDone + ((dwDiff > 8) ? 8 : dwDiff);

		byteTmpBuff[dwBuffIdx] = 0;
		do
		{
			if (IsPartShared(dwDone++))
				byteTmpBuff[dwBuffIdx] |= byteMask;
			byteMask <<= 1;
		} while (dwDone < dwLimit);
		dwBuffIdx = (dwBuffIdx + 1) & (sizeof(byteTmpBuff) - 1);
		if (dwBuffIdx == 0)
			file->Write(byteTmpBuff, sizeof(byteTmpBuff));
	}
	if (dwBuffIdx != 0)
		file->Write(byteTmpBuff, dwBuffIdx);

	EMULE_CATCH
}
//--- :xrmb/partprio ---
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// obaldin
bool CKnownFile::IsJsComplete()
{
#ifndef NEW_SOCKETS_ENGINE
     return CJumpstarter::IsJsCompleteForFile(this);
#else
	return true;
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::SetJumpstartEnabled(bool enabled)
{
#ifndef NEW_SOCKETS_ENGINE
    if(!enabled && m_Jumpstarter)
	{
        m_Jumpstarter->Disable();
        delete m_Jumpstarter;
        m_Jumpstarter = NULL;
    }
	else if(enabled && !m_Jumpstarter)
	{
        CJumpstarter::EnableForFile(this);
        m_Jumpstarter = new CJumpstarter(this);
    }
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::AddSentBlock(CUpDownClient* client,uint32 start_offset, uint32 togo)
{
#ifndef NEW_SOCKETS_ENGINE
    m_Jumpstarter->AddSentBlock(client, start_offset, togo);
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CKnownFile::WriteJumpstartPartStatus(CUpDownClient* client, CMemFile* data)
{
#ifndef NEW_SOCKETS_ENGINE
    m_Jumpstarter->WriteJumpstartPartStatus(client, data);
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CKnownFile::AllowChunkForClient(uint32 partNo, CUpDownClient* client)
{
#ifndef NEW_SOCKETS_ENGINE
    return (!m_Jumpstarter) || m_Jumpstarter->AllowChunkForClient(partNo, client);
#else
	return true;
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	ReadFileForUpload() opens known file and read information in circular way.
//		Params:
//			dwOffset      - file offset to start reading from;
//			dwBytesToRead - number of bytes to read;
//			pbyteBuffer   - buffer to receive data.
//		Return:
//			< 0 in case of error, else 0.
int CKnownFile::ReadFileForUpload(uint32 dwOffset, uint32 dwBytesToRead, byte *pbyteBuffer)
{
	CString	strFullName;

	strFullName.Format(_T("%s\\%s"), m_strKnownFileDirectory, m_strFileName);

	return (::CircularFileRead(strFullName, dwOffset, dwBytesToRead, pbyteBuffer)) ? 0 : -1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CKnownFile::GetPermissionString()
{
#ifndef NEW_SOCKETS_ENGINE
	static const uint16 s_auResTbl[] =
	{
		IDS_FSTATUS_PUBLIC,			//PERM_ALL
		IDS_FSTATUS_FRIENDSONLY,	//PERM_FRIENDS
		IDS_HIDDEN					//PERM_NOONE
	};
	unsigned	uiPerm = GetPermissions();

	if (uiPerm >= ELEMENT_COUNT(s_auResTbl))
		uiPerm = PERM_ALL;
	return GetResString(s_auResTbl[uiPerm]);
#else
	return _T("");
#endif //NEW_SOCKETS_ENGINE
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
