//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "zlib/zlib.h"
#include "UpDownClient.h"
#include "opcodes.h"
#include "packets.h"
#include "emule.h"
#include "UploadQueue.h"
#include "otherstructs.h"
#include "otherfunctions.h"
#include "math.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#define new DEBUG_NEW
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::DrawUpStatusBar(CDC *pDC, RECT *pRect, bool bFlat)
{
	EMULE_TRY

	const COLORREF	crSentData = RGB(222, 160, 0);
	const COLORREF	crSendingPart = RGB(255, 208, 0);
	const COLORREF	crMeOnly = RGB(190, 190, 255);
	COLORREF		crBoth, crNeither, crClientOnly;

	if (bFlat)
	{
		crBoth = RGB(0, 150, 0);
		crNeither = RGB(224, 224, 224);
		crClientOnly = RGB(0, 0, 0);
	}
	else
	{
		crBoth = RGB(0, 192, 0);
		crNeither = RGB(240, 240, 240);
		crClientOnly = RGB(95, 95, 95);
	}

//	Get and set the file's size
	CPartFile	   *pUploadingFile = reinterpret_cast<CPartFile*>(g_eMuleApp.m_pSharedFilesList->GetFileByID(m_reqFileHash));
	uint32			dwFileSize = (pUploadingFile != NULL) ? pUploadingFile->GetFileSize() : 1;
	CBarShader		statusBar(pRect->bottom - pRect->top, pRect->right - pRect->left, crNeither, dwFileSize);

	if ((m_pbyteUpPartStatuses != NULL) && (pUploadingFile != NULL))
	{
		uint32		dwUploadingPart = GetCurrentlyUploadingPart();
		uint32		dwStartPos, dwEndPos;

		for (uint32 i = 0; i < m_uUpPartCount; i++)
		{
			dwStartPos = PARTSIZE * i;
			dwEndPos = ((dwFileSize - dwStartPos) > PARTSIZE) ? (dwStartPos + PARTSIZE) : dwFileSize;
			if (i == dwUploadingPart)
			{
			//	Add the current uploading part
				statusBar.FillRange(dwStartPos, dwEndPos, crSendingPart);

				ReqBlockDeque::iterator	itBlock;
				Requested_Block_Struct	*pDoneBlock;

			//	Add already uploaded pieces
				for (itBlock = m_doneBlocksList.begin(); itBlock != m_doneBlocksList.end(); itBlock++)
				{
					pDoneBlock = *itBlock;
					if (dwUploadingPart == (pDoneBlock->m_dwStartOffset / PARTSIZE))
						statusBar.FillRange(pDoneBlock->m_dwStartOffset, pDoneBlock->m_dwEndOffset, crSentData);
				}
			}
			else if (!pUploadingFile->IsPartFile() || pUploadingFile->IsPartComplete(i))	// We've this part
			{
				statusBar.FillRange(dwStartPos, dwEndPos, (m_pbyteUpPartStatuses[i] != 0) ? crBoth : crMeOnly);
			}
			else if (m_pbyteUpPartStatuses[i] != 0)	// The remote user has this part
			{
				statusBar.FillRange(dwStartPos, dwEndPos, crClientOnly);
			}
		}
	}
	statusBar.Draw(pDC, pRect->left, pRect->top, bFlat);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::IsCommunity()
{
	EMULE_TRY
	
	if (m_byteActionsOnNameChange & AONC_COMMUNITY_CHECK)
	{
		m_bIsCommunity = false;

		if (g_eMuleApp.m_pGlobPrefs->CommunityEnabled())
		{
			CString		strLowerCommunity = g_eMuleApp.m_pGlobPrefs->CommunityString();

			strLowerCommunity.MakeLower();
			strLowerCommunity.Remove(_T('['));
			strLowerCommunity.Remove(_T(']'));

			if (!strLowerCommunity.IsEmpty())
			{
				CString		strProperCommunity;

				strProperCommunity.Format(_T("[%s]"), strLowerCommunity);

			//	Ignore unsuitable tags which are not meant to be used as community
				CString		strLowerUserName = m_strUserName;

				strLowerUserName.MakeLower();
				strLowerUserName.Replace(_T("[eplus]"), _T(""));

			//	Only the first community-tag is evaluated to be valid
				int		iFirstTagPos = strLowerUserName.Find(_T('['));

				if (iFirstTagPos >= 0 && iFirstTagPos == strLowerUserName.Find(strProperCommunity))
					m_bIsCommunity = true;
			}
		}

		m_byteActionsOnNameChange &= ~AONC_COMMUNITY_CHECK;
	}

	return m_bIsCommunity;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::GetScore(bool bIsDownloading, bool bOnlyBaseValue /*= false*/)
{
	static DWORD	dwLastMsgTime = 0;

	EMULE_TRY

	if (m_strUserName.IsEmpty())
		return 0;

	if (m_bBanned)
	{
		return 0;
	}
	else
	{
		if ( HasUserNameForbiddenStrings() || HasMODNameForbiddenStrings()
			 || ((GetModString() == _T(PLUS_VERSION_STR)) && (GetPlusVersion() == 0)) )
		{
		//	Spam control
			if (::GetTickCount() > dwLastMsgTime + (60 * 1000))
			{
				AddDebugLogLine(RGB_LOG_DIMMED_TXT _T("Anti-leechermods: Client '%s' (%s) has been banned"), GetUserName(), GetClientNameAndVersionString());

				Ban(_T("AUTOMATED MESSAGE: You're being banned because your client is leeching!"), GetResString(IDS_BAN_LEECHERADV));

				dwLastMsgTime = ::GetTickCount();
			}
			return 0;
		}
	}

#ifdef OLD_SOCKETS_ENABLED
	if (!bIsDownloading && !bOnlyBaseValue)
	{
		if (HasLowID() && !(m_pRequestSocket && m_pRequestSocket->IsConnected()))
		{
			if (!g_eMuleApp.m_pServerConnect->IsConnected() || g_eMuleApp.m_pServerConnect->IsLowID())
				return 0;
		}
	}
#endif //OLD_SOCKETS_ENABLED

	CKnownFile* pSharedFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(m_reqFileHash);

	if (pSharedFile == NULL)
		return 0;

	ASSERT(m_reqFileHash != NULL);

	double	dblFilePriority;

	switch (pSharedFile->GetULPriority())
	{
		case PR_RELEASE:
			if (pSharedFile->IsPartFile())
				dblFilePriority = 5.0;
			else
				dblFilePriority = 10.0;
			break;

		case PR_HIGH:
			dblFilePriority = 2.0;
			break;

		case PR_LOW:
			dblFilePriority = 0.5;
			break;

		case PR_VERYLOW:
			dblFilePriority = 0.2;
			break;

		case PR_NORMAL:
		default:
			dblFilePriority = 1.0;	// standard
			break;
	}

	double	dblSizeRatio = pSharedFile->GetSizeRatio();
	double	dblPopularityRatio = pSharedFile->GetPopularityRatio();

//	Calculate score, based on waiting time and other factors
	double		dblBaseValue;
	if (bOnlyBaseValue)
	{
		dblBaseValue = 100.0f;
	}
	else if (!bIsDownloading)
	{
		dblBaseValue = static_cast<double>(GetTickCount() - m_dwWaitTime) / 1000.0;
	}
	else
	{
	//	We don't want one client to download forever
	//	the first 15 min download time counts as 15 min waitingtime and you get a 15 min bonus while you are in the first 15 min :)
	//	(to avoid 20 sec downloads) after this the score won't raise anymore
		dblBaseValue = static_cast<double>(m_dwUploadTime - m_dwWaitTime);
		ASSERT (m_dwUploadTime - m_dwWaitTime >= 0); //oct 28, 02: changed this from "> 0" to ">= 0"
		dblBaseValue += static_cast<double>((GetTickCount() - m_dwUploadTime > 900000) ? 900000.0 : 1800000.0);
		dblBaseValue /= 1000.0;
	}

	if (IsFriend())
	{
	//	Client is on friendlist
		dblBaseValue *= 2.0;
	}
	else
	{
	//	Client is a community user
		if (IsCommunity())
			dblBaseValue *= 1.5;
	}

//	Client has credits (modifier between 1.0 and 10.0)
	if (m_pCredits != NULL)
		dblBaseValue *= m_pCredits->GetScoreRatio(this->GetIP());

//	Little penalty for _very_ old clients that are harmful for servers and network (from official)
	if((IsEmuleClient() && m_byteEmuleVersion <= 0x19) || this->GetClientSoft() == SO_SHAREAZA)
		dblBaseValue *= 0.5f;

	dblBaseValue *= dblFilePriority * dblSizeRatio * dblPopularityRatio;

	return static_cast<uint32>(dblBaseValue);

	EMULE_CATCH

	return 1;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Checks if it is next requested block from another chunk of the actual file or from another file
//	[Returns]
//	true : Next requested block is from another different chunk or file than last downloaded block
//	false: Next requested block is from same chunk that last downloaded block
bool CUpDownClient::IsDifferentPartBlock(bool bSilent /*= false*/)
{
	bool	different_part = false;

	try
	{
	//	Check if we have good lists and proceed to check for different chunks
		if (!m_blockRequestsQueue.empty() && !m_doneBlocksList.empty())
		{
		//	Get last block and next pending
			Requested_Block_Struct	*last_done_block = m_doneBlocksList.front();
			Requested_Block_Struct	*next_requested_block = m_blockRequestsQueue.front();

		//	Calculate corresponding parts to blocks
			uint32	last_done_part = last_done_block->m_dwStartOffset / PARTSIZE;
			uint32	next_requested_part = next_requested_block->m_dwStartOffset / PARTSIZE;

		//	Test is we are asking same file and same part
			if (last_done_part != next_requested_part || md4cmp(last_done_block->m_fileHash, next_requested_block->m_fileHash) != 0)
				different_part = true;
		}
	}
	catch (...)
	{
		different_part = true;
	}

	if (different_part)
	{
	//	obaldin: this is for old clients, that rotate the chunk, so that they'll be able to download at least _something_
	//	But probably even after the majority switches we'd better check for 0 here
		if (GetSessionUp() < 1024 * 1024)
		{
			different_part = false;

			if (!bSilent && g_eMuleApp.m_pGlobPrefs->IsClientTransferLogEnabled())
				AddDebugLogLine(_T("Client '%s' (%s) continues downloading (should be back on queue) because of a low transferred size"), GetUserName(), GetClientNameAndVersionString());
		}
	}

	return different_part;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::IsForbiddenByJumpstart()
{
	EMULE_TRY

	Requested_Block_Struct * next_requested_block;
	uint32 next_requested_part = 0xffffffff;

	if (m_blockRequestsQueue.empty())
	{
		return false;
	}
	else
	{
		next_requested_block = (Requested_Block_Struct*)m_blockRequestsQueue.front();
		next_requested_part = next_requested_block->m_dwStartOffset / PARTSIZE;
		CKnownFile* kf = g_eMuleApp.m_pSharedFilesList->GetFileByID(next_requested_block->m_fileHash);

		if (kf != NULL && !kf->AllowChunkForClient(next_requested_part, this))
			return true;
	}

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
class CSyncHelper
{
public:

	CSyncHelper()
	{
		m_pObject = NULL;
	}

	~CSyncHelper()
	{
		if (m_pObject)
			m_pObject->Unlock();
	}

	CSyncObject* m_pObject;
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	Check if we should kick this client (time critical !!!)
//	- Always kick if transfer is longer than 80 minutes
//	- Never kick friends before a chunck was transferred
//	- Never allow more than the size of a chunk (+500 KBytes for small resends of invalid data) to be transferred
//	- We never kick a lancast user
//	- Full chunk transfers are being enforced, is not more an option
//
EnumBlockPacketCreationResult CUpDownClient::CreateNextBlockPackage()
{
	EMULE_TRY

//	If there are no new blocks requested, we will happily transfer more if the client request it
	if (m_blockRequestsQueue.empty())
		return BPCR_FAILED_NO_REQUESTED_BLOCKS;

	bool bAllowedMore = IsOnLAN();
	bool bDifferentPart = IsDifferentPartBlock(true);
	bool bJSIsBlocked = IsForbiddenByJumpstart();
	bool bShouldContinue = ((GetUpStartTimeDelay() >= 80 * 60 * 1000) || bDifferentPart);

//	If user is from Internet & Full chunk (or timeout)
	if (!bAllowedMore && bShouldContinue)
	{
		if ( IsFriend() && GetSessionUp() < (PARTSIZE + EMBLOCKSIZE * 6)
			 && !g_eMuleApp.m_pUploadQueue->CheckForTimeOver(this) && !bJSIsBlocked )
		{
			bAllowedMore = true;
			if (g_eMuleApp.m_pGlobPrefs->IsClientTransferLogEnabled())
				AddDebugLogLine(_T("Client '%s' (%s) downloaded %u KB so far. Allowed to continue downloading (should go back on queue) because it's a friend"), GetUserName(), GetClientNameAndVersionString(), this->GetSessionUp() / 1024);
		}
		else
		{
			if (bDifferentPart)
				return BPCR_FAILED_BLOCKED_PART;
			else if (bJSIsBlocked)
				return BPCR_FAILED_BLOCKED_PART;
			else
				return BPCR_FAILED_TIME_LIMIT;
		}
	}

	CFile			file;
	CKnownFile*		srcfile;
	byte*			filedata = 0;
	EnumBlockPacketCreationResult eBPCResult = BPCR_FAILED_TIME_LIMIT;

	try
	{
	//	Note: since this function is called when upload packets queue is empty,
	//	the upload can be broken during packets creation for requested blocks,
	//	therefore in order to decrease the probability of connection loss,
	//	we gonna create just one requested block packet.
	//	This will lead to
	//		1) faster response after initial request -> better number of successful UL sessions
	//		2) reduced delay between some packets -> smoother upload -> less possibility of connection loss
		if ( (bAllowedMore || !IsDifferentPartBlock()) && !IsForbiddenByJumpstart() )
		{
			Requested_Block_Struct	   *pCurrentBlock = m_blockRequestsQueue.front();

			srcfile = g_eMuleApp.m_pSharedFilesList->GetFileByID(pCurrentBlock->m_fileHash);

			if (srcfile == NULL)
			{
				CString		strError;

				eBPCResult = BPCR_FAILED_FILE_ERROR;
				strError.Format(_T("requested file not found, hash: %s"), HashToString(pCurrentBlock->m_fileHash));
				throw strError;
			}

		//	Update the file hash to the requested file
			md4cpy(m_reqFileHash, pCurrentBlock->m_fileHash);

		//	Check that block start is within the file
			if (pCurrentBlock->m_dwStartOffset >= srcfile->GetFileSize())
			{
				CString		strError;

				eBPCResult = BPCR_FAILED_FILE_ERROR;
				strError.Format( _T("request data %#x-%#x beyond the file end %#x '%s'"),
									pCurrentBlock->m_dwStartOffset, pCurrentBlock->m_dwEndOffset, srcfile->GetFileSize(), srcfile->GetFileName());
				throw strError;
			}

			uint32 dwPart = pCurrentBlock->m_dwStartOffset / PARTSIZE;

		//	Accessing a blocked part ?
			if (!srcfile->GetJumpstartEnabled() && !srcfile->IsPartShared(dwPart))
			{
				CString		strError;

				AddLogLine(false,  IDS_PARTPRIO_DENIED, dwPart, srcfile->GetFileName());
				strError.Format(GetResString(IDS_PARTPRIO_DENIED), dwPart, srcfile->GetFileName());
				eBPCResult = BPCR_FAILED_BLOCKED_PART;
				throw strError;
			}

			uint32 dwToGo;
			uint32 dwStart = dwPart * PARTSIZE;
			uint32 dwEnd = ((srcfile->GetFileSize() - dwStart) > PARTSIZE) ? (dwStart + PARTSIZE) : srcfile->GetFileSize();

		//	Everyone is limited to a single chunk.
		//	Check that m_dwStartOffset and m_dwEndOffset are in the same chunk and
		//	make sure we don't pass the end of the file.
			if ( (dwPart != ((pCurrentBlock->m_dwEndOffset - 1) / PARTSIZE)) ||
				(pCurrentBlock->m_dwEndOffset > dwEnd) )
			{
			//	m_dwEndOffset goes into the next chunk or is beyond the file end.
			//	Set it to the end of the chunk that m_dwStartOffset is in.
				pCurrentBlock->m_dwEndOffset = dwEnd;
			}

		//	This can't be a wrapped around request, since it has been limited to a single chunk.
			dwToGo = pCurrentBlock->m_dwEndOffset - pCurrentBlock->m_dwStartOffset;

			if (srcfile->IsPartFile())
			{
				if (!(reinterpret_cast<CPartFile*>(srcfile))->IsPartComplete(dwPart))
				{
					CString		strError;

					strError.Format(_T("asked for incomplete block from '%s'"), srcfile->GetFileName());
					eBPCResult = BPCR_FAILED_BLOCKED_PART;
					throw strError;
				}
			}

		//	Create a buffer for file data before we start to work with the file
			filedata = new byte[dwToGo + 500];

			int	iRc = srcfile->ReadFileForUpload(pCurrentBlock->m_dwStartOffset, dwToGo, filedata);

			if (iRc < 0)
			{
			//	Error occurred
				CString		strError;

				strError.Format(_T("failed to open '%s'"), srcfile->GetFileName());
				eBPCResult = BPCR_FAILED_FILE_ERROR;
				throw strError;
			}

			md4cpy(m_reqFileHash, pCurrentBlock->m_fileHash);

			bool bFromPF = (iRc) ? true : false;
			bool bPackedCreated = false;
			bool bCompress = false;

			if (m_byteDataCompVer == 1)
			{
				bCompress = srcfile->IsCompressedTransferAllowed();
			}

			if (bCompress)
			{
				bPackedCreated = CreatePackedPackets(filedata, dwToGo, pCurrentBlock, srcfile->GetULPriority(), bFromPF);
			}
			else
			{
				bPackedCreated = CreateStandardPackets(filedata, dwToGo, pCurrentBlock, srcfile->GetULPriority(), bFromPF);
			}

		//	Update the data only if packet was created
			if (bPackedCreated)
			{
			//	File statistic
				srcfile->statistic.AddTransferred(dwToGo);

				srcfile->statistic.AddTraffic(pCurrentBlock->m_dwStartOffset, dwToGo);

				if (srcfile->GetJumpstartEnabled())
				{
					srcfile->AddSentBlock(this, pCurrentBlock->m_dwStartOffset, dwToGo);
				}

				m_doneBlocksList.push_front(m_blockRequestsQueue.front());

				eBPCResult = BPCR_OK;
			}

			m_blockRequestsQueue.pop_front();
			delete[] filedata;
			filedata = 0;
		}
	}
//	Check for exception that will be throw by "new"
	catch (CMemoryException * error)
	{
		AddDebugLogLine(RGB_LOG_ERROR + _T("Failed to create upload package for %s - %s"), GetUserName(), GetErrorMessage(error));
		error->Delete();
		SetUploadFileID(NULL);
		return BPCR_FAILED_FILE_ERROR;
	}
	catch (CFileException * error)
	{
		OUTPUT_DEBUG_TRACE();
		AddDebugLogLine(RGB_LOG_ERROR + _T("Failed to create upload package for %s - %s"), GetUserName(), GetErrorMessage(error));
		SetUploadFileID(NULL);
		if (filedata != NULL)
			delete[] filedata;
		error->Delete();
		return BPCR_FAILED_FILE_ERROR;
	}
	catch (CString &strError)
	{
		OUTPUT_DEBUG_TRACE();
		AddDebugLogLine(_T("Client '%s' (%s) caused an error while creating package (%s), disconnecting!"), GetUserName(), GetClientNameAndVersionString(), strError);
		SetUploadFileID(NULL);
		if (filedata != NULL)
			delete[] filedata;
		return eBPCResult;
	}

	return eBPCResult;

	EMULE_CATCH

	return BPCR_FAILED_FILE_ERROR;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::CheckForUpPartsInfo()
{
	EMULE_TRY

	if ( m_uPartCount == 0 || m_pReqPartFile == NULL
		|| md4cmp(m_reqFileHash, m_pReqPartFile->GetFileHash())
		|| m_pbytePartStatuses == NULL )
	{
		return;
	}

	if (m_pbyteUpPartStatuses)
	{
		delete[] m_pbyteUpPartStatuses;
		m_pbyteUpPartStatuses = NULL;
	}

	m_uUpPartCount = m_uPartCount;
	m_uAvailUpPartCount = 0;
	m_pbyteUpPartStatuses = new byte[m_uUpPartCount];

	for (int i = 0; i < m_uUpPartCount; i++)
	{
		m_pbyteUpPartStatuses[i] = m_pbytePartStatuses[i];

		if (m_pbyteUpPartStatuses[i] != NULL)
			m_uAvailUpPartCount++;
	}
	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::ProcessUpFileStatus(char *pcPacket, uint32 dwPacketSize)
{
	EMULE_TRY

	if (m_pbyteUpPartStatuses)
	{
		delete[] m_pbyteUpPartStatuses;
		m_pbyteUpPartStatuses = NULL;
	}
	m_uUpPartCount = 0;
	m_uAvailUpPartCount = 0;
	m_uUpCompleteSourcesCount = 0;
	if (dwPacketSize == 16)
	{
		CheckForUpPartsInfo();
		return;
	}

	CSafeMemFile	packetStream(reinterpret_cast<BYTE *>(pcPacket), dwPacketSize);
	uchar			abyteFileHash[16];

	packetStream.Read(abyteFileHash, 16);

	CKnownFile	   *pKnownFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(abyteFileHash);
	uint16			uED2KUpPartCount;

	packetStream.Read(&uED2KUpPartCount, 2);
	m_uUpPartCount = pKnownFile->GetPartCount();

	if (uED2KUpPartCount == 0)
	{
		m_pbyteUpPartStatuses = new byte[m_uUpPartCount];
		memzero(m_pbyteUpPartStatuses, m_uUpPartCount);
	}
	else
	{
		if (pKnownFile->GetED2KPartCount() != uED2KUpPartCount)
		{
			m_uUpPartCount = 0;
			return;
		}

		m_pbyteUpPartStatuses = new byte[m_uUpPartCount];

		uint16	uDone = 0;

		while (uDone != m_uUpPartCount)
		{
			byte	byteToRead;

			packetStream.Read(&byteToRead, 1);

			for (sint32 i = 0;i != 8;i++)
			{
				m_pbyteUpPartStatuses[uDone] = ((byteToRead >> i) & 1) ? 1 : 0;

				if (m_pbyteUpPartStatuses[uDone])
					m_uAvailUpPartCount++;
				uDone++;
				if (uDone == m_uUpPartCount)
					break;
			}
		}

		if ((GetExtendedRequestsVersion() > 1) && (dwPacketSize >= 20 + (((uint32)m_uUpPartCount + 7) >> 3)))
		{
			uint16	uCompleteCountLast = GetUpCompleteSourcesCount();
			uint16	uCompleteCountNew;

			packetStream.Read(&uCompleteCountNew, 2);
			SetUpCompleteSourcesCount(uCompleteCountNew);
			if (uCompleteCountLast != uCompleteCountNew)
				pKnownFile->CalculateCompleteSources();
		}
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::CreateStandardPackets(byte *pbyteFileData, uint32 dwToGo, Requested_Block_Struct *pCurrentBlock, byte byteFilePriority, bool bFromPF /*= true*/)
{
	EMULE_TRY

	uint32			dwPacketSize;
	CMemFile		memfile(pbyteFileData, dwToGo);
	PacketDeque		blockSendTempQueue;

	if (dwToGo > 10240)
		dwPacketSize = dwToGo / (dwToGo / 10240L);
	else
		dwPacketSize = dwToGo;

	while (dwToGo != 0)
	{
		if (dwToGo < dwPacketSize * 2)
			dwPacketSize = dwToGo;

		dwToGo -= dwPacketSize;

		Packet		*pPacket = new Packet(OP_SENDINGPART, dwPacketSize + 24, OP_EDONKEYPROT, byteFilePriority, bFromPF);

		md4cpy(&pPacket->m_pcBuffer[0], m_reqFileHash);

		uint32 startpos = (pCurrentBlock->m_dwEndOffset - dwToGo) - dwPacketSize;
		uint32 endpos = (pCurrentBlock->m_dwEndOffset - dwToGo);

		memcpy2(&pPacket->m_pcBuffer[16], &startpos, 4);
		memcpy2(&pPacket->m_pcBuffer[20], &endpos, 4);

	//	Append the file data to the packet
		memfile.Read(&pPacket->m_pcBuffer[24], dwPacketSize);
	//	If the split was successful, then move the splited packets to global queue
		if (SplitAndAddPacketsToSendQueue(pPacket, &blockSendTempQueue, byteFilePriority, bFromPF))
		{
		//	Attach splitted packets to global queue
			while (!blockSendTempQueue.empty())
			{
				m_blockSendQueue.push_back(blockSendTempQueue.front());
				blockSendTempQueue.pop_front();
			}
		}
		else
		{
			while (!blockSendTempQueue.empty())
			{
				delete blockSendTempQueue.front();
				blockSendTempQueue.pop_front();
			}
		}
	}

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CUpDownClient::CreatePackedPackets(byte *pbyteFileData, uint32 dwToGo, Requested_Block_Struct *pCurrentBlock, byte byteFilePriority, bool bFromPF /*= true*/)
{
	EMULE_TRY

	BYTE * output = new BYTE[dwToGo + 300];
	uLongf newsize = dwToGo + 300;
	uint16 result = compress2(output, &newsize, pbyteFileData, dwToGo, 9);

	if (result != Z_OK || dwToGo <= newsize)
	{
		delete[] output;
		return CreateStandardPackets(pbyteFileData, dwToGo, pCurrentBlock, byteFilePriority, bFromPF);;
	}

	CMemFile		memfile(output, newsize);
	PacketDeque		blockSendTempQueue;
	uint32			dwPacketSize;

//	Update the compression statistic
	m_dwCompressionGain += (dwToGo - newsize);
	m_dwUncompressed += dwToGo;

	dwToGo = newsize;
	if (dwToGo > 10240)
		dwPacketSize = dwToGo / (uint32)(dwToGo / 10240);
	else
		dwPacketSize = dwToGo;

	while (dwToGo != 0)
	{
		if (dwToGo < dwPacketSize * 2)
			dwPacketSize = dwToGo;

		dwToGo -= dwPacketSize;

		Packet	   *pPacket = new Packet(OP_COMPRESSEDPART, dwPacketSize + 24, OP_EMULEPROT, byteFilePriority, bFromPF);
		uint32		startpos = pCurrentBlock->m_dwStartOffset;

		md4cpy(&pPacket->m_pcBuffer[0], m_reqFileHash);
		memcpy2(&pPacket->m_pcBuffer[16], &startpos, 4);
		memcpy2(&pPacket->m_pcBuffer[20], &newsize, 4);
		memfile.Read(&pPacket->m_pcBuffer[24], dwPacketSize);

	//	If one of the split action was failed, then cancel whole packet creation, cause otherwise Z-stream on other side will be corrupted
		if (!SplitAndAddPacketsToSendQueue(pPacket, &blockSendTempQueue, byteFilePriority, bFromPF))
		{
			while (!blockSendTempQueue.empty())
			{
				delete blockSendTempQueue.front();
				blockSendTempQueue.pop_front();
			}
			delete[] output;
			return false;
		}
	}
	delete[] output;

//	Attach splitted packets to global queue
	while (!blockSendTempQueue.empty())
	{
		m_blockSendQueue.push_back(blockSendTempQueue.front());
		blockSendTempQueue.pop_front();
	}

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// A packet should never be larger than MAXFRAGSIZE. This method splits up the
// given packets into several packets with the size MAXFRAGSIZE, if necessary.
bool CUpDownClient::SplitAndAddPacketsToSendQueue(Packet *pPacket, PacketDeque *pSendQueue, byte byteFilePriority, bool bFromPF)
{
	EMULE_TRY

	if (pPacket->GetRealPacketSize() > 0 && pPacket->GetRealPacketSize() > MAXFRAGSIZE * 2)
	{
	//	Splitting packets
		uint32		nSize = pPacket->GetRealPacketSize();
		char	   *m_pcBuffer = pPacket->DetachPacket();

		delete pPacket;

		uint32	nPos = 0;
		int		iAllocationTries = 0;

		while (nPos < nSize && iAllocationTries < 3)
		{
			char	   *pBuffer2 = NULL;
			Packet	   *pPacket = NULL;
			uint32		nNewSize = (nSize - nPos < MAXFRAGSIZE) ? nSize - nPos : MAXFRAGSIZE;

			try
			{
				pBuffer2 = new char[nNewSize];
				pPacket = new Packet(pBuffer2, nNewSize, (nNewSize < MAXFRAGSIZE), byteFilePriority, bFromPF);
			//	To optimize handling in case of failed packet allocation, let's copy the data after packet creation
				memcpy2(pBuffer2, m_pcBuffer + nPos, nNewSize);
			//	Put packet in temporary queue
				pSendQueue->push_back(pPacket);
			//	Update the data pointer for next fragment only if packet allocation was successful
			//	in this case we can retry to allocate the packet again
				nPos += nNewSize;
				iAllocationTries = 0;
			}
			catch (CMemoryException *error)
			{
				error->Delete();
				safe_delete(pBuffer2);
				iAllocationTries++;
			}

		}
		delete[] m_pcBuffer;

	//	If the split was failed due to problem with memory allocation report it back
		if (iAllocationTries != 0)
			return false;
	}
	else
	{
		m_blockSendQueue.push_back(pPacket);
	}

	return true;

	EMULE_CATCH

	return false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetUploadFileID(uchar *pbyteReqFileID)
{
	EMULE_TRY

	if (pbyteReqFileID != NULL)
		md4cpy(m_reqFileHash, pbyteReqFileID);
	else
		md4clr(m_reqFileHash);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::AddReqBlock(Requested_Block_Struct *pReqBlock)
{
	EMULE_TRY

	ReqBlockDeque::iterator		itBlock;
	Requested_Block_Struct	   *pTempBlock;

	for (itBlock = m_doneBlocksList.begin(); itBlock != m_doneBlocksList.end(); itBlock++)
	{
		pTempBlock = *itBlock;

		if (pTempBlock != NULL && pReqBlock->m_dwStartOffset == pTempBlock->m_dwStartOffset
			&& pReqBlock->m_dwEndOffset == pTempBlock->m_dwEndOffset)
		{
			return;
		}
	}

	for (itBlock = m_blockRequestsQueue.begin(); itBlock != m_blockRequestsQueue.end(); itBlock++)
	{
		pTempBlock = *itBlock;

		if (pTempBlock != NULL && pReqBlock->m_dwStartOffset == pTempBlock->m_dwStartOffset
			&& pReqBlock->m_dwEndOffset == pTempBlock->m_dwEndOffset)
		{
			return;
		}
	}

	pTempBlock = new Requested_Block_Struct;

	memcpy2(pTempBlock, pReqBlock, sizeof(Requested_Block_Struct));
	m_blockRequestsQueue.push_back(pTempBlock);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetUpStartTime(uint32 dwTime)
{
	EMULE_TRY

	if (dwTime != 0)
		m_dwUploadTime = dwTime;
	else
		m_dwUploadTime = ::GetTickCount();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetWaitStartTime(uint32 dwTime)
{
	EMULE_TRY

	if (dwTime != 0)
		m_dwWaitTime = dwTime;
	else
		m_dwWaitTime = ::GetTickCount();

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::SendBlockData(uint32 dwMaxAmount, EnumBlockSendResult& eResult)
{
	eResult = BSR_FAILED_CLIENT_LEFT_UQ;

	uint32	dwTimePeriod, dwAmountTransferred = 0;
	DWORD	curTick = GetTickCount();

	try
	{
	//	Check for 'stuck' clients
		if ((curTick - GetLastGotULData() > 2 * 60 * 1000) && (GetUpStartTimeDelay() > 2 * 60 * 1000))
		{
		//	We were unable to transfer for 2 minutes, just remove the client from UL queue
		//	If client stays online, he will request file & be added to the queue
			if (g_eMuleApp.m_pGlobPrefs->IsClientTransferLogEnabled())
				AddDebugLogLine( _T("Client '%s' (%s) appears to be stuck, putting back on queue"),
								 GetUserName(), GetClientNameAndVersionString() );
			g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_TIMEOUT);
			return 0;
		}

#ifdef OLD_SOCKETS_ENABLED
	//	Only feed this slot new data if it has been able to empty its buffer since we last offered data
		if (m_pRequestSocket != NULL)
		{
			if (m_pRequestSocket->IsBusy())
			{
			//	Check block queues to separate the active client from just invited to upload
				if (m_blockRequestsQueue.empty() && m_blockSendQueue.empty())
					eResult = BSR_FAILED_NO_REQUESTED_BLOCKS;
				else
					eResult = BSR_BUSY;
			}
			else
			{
			//	1) Measure only already transferred data! the speed should be calculated before data transfer, otherwise
			//		not yet transferred data from current iteration will lead to higher(incorrect) speed value
			//	2) Use socket status to find out the correct time for the speed measurements, because any calculation of the average inbetween
			//		will decrease the calculated value, what does not reflect real situation
				m_averageUDRList.push_front(m_dwTransferredUp);
				m_averageULTickList.push_front(curTick);

			//	Don't store too much statistics. maximal 40 sec ~ socket timeout
				while ((dwTimePeriod = (curTick - m_averageULTickList.back())) > CONNECTION_TIMEOUT)
				{
					m_averageUDRList.pop_back();
					m_averageULTickList.pop_back();
				}

			//	Calculate average data rate
				sint32 lCurUpDataRate = -1;

			//	Leave this slot's speed as 0 unless there's a good chunk of stored statistics for it (at least 10 samples ~ 1 sec).
				if (dwTimePeriod > 1000)
				{
					lCurUpDataRate = static_cast<sint32>((static_cast<double>(m_averageUDRList.front() - m_averageUDRList.back())) * 1000.0 / dwTimePeriod);
				}

			//	Assign & update only if values are different
				if (m_lUpDataRate != lCurUpDataRate)
				{
					m_lUpDataRate = lCurUpDataRate;
					g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlUploadList.UpdateClient(this);
				}

				if (GetUploadState() == US_UPLOADING)
				{
					if (m_blockSendQueue.empty())
					{
					// If all packets were sent, try to create a new one
						EnumBlockPacketCreationResult	eBPCResult = CreateNextBlockPackage();

						if (eBPCResult != BPCR_OK)
						{
						//	If packets creation was failed, client can be removed without data loss
							if (eBPCResult == BPCR_FAILED_NO_REQUESTED_BLOCKS)
							{
							//	If client did not request data, keep him in the queue & give him a chance to request something
							//	The client will be removed from UL queue after 2 min.
								eResult = BSR_FAILED_NO_REQUESTED_BLOCKS;
								return 0;
							}
							else if (eBPCResult == BPCR_FAILED_BLOCKED_PART)
							{
								if (g_eMuleApp.m_pGlobPrefs->IsClientTransferLogEnabled())
									AddDebugLogLine(_T("Client %s (%s) left UL queue after receiving %u KB. Asked for new or blocked chunk/file"), GetUserName(), GetClientNameAndVersionString(), GetSessionUp() / 1024);
								g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BLOCKED_CHUNK);
							}
							else if (eBPCResult == BPCR_FAILED_FILE_ERROR)
							{
								if (g_eMuleApp.m_pGlobPrefs->IsClientTransferLogEnabled())
									AddDebugLogLine(_T("Client %s (%s) left UL queue after receiving %u KB. Asked for unaccessible data (file removed, wrong request, etc)"), GetUserName(), GetClientNameAndVersionString(), GetSessionUp() / 1024);
								g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_FILE_ERROR);
							}
							else
							{
								if (g_eMuleApp.m_pGlobPrefs->IsClientTransferLogEnabled())
									AddDebugLogLine(_T("Client %s (%s) left UL queue after receiving %u KB. Upload timeout (low speed, no block request)"), GetUserName(), GetClientNameAndVersionString(), GetSessionUp() / 1024);
								g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_TIMEOUT);
							}

						//	Notify the remote client about the end of the session
						//	Note (offical): 
						//		OP_OUTOFPARTREQS will tell the downloading client to go back to OnQueue..
						//		The main reason for this is that if we put the client back on queue and it goes
						//		back to the upload before the socket times out... We get a situation where the
						//		downloader thinks it already sent the requested blocks and the uploader thinks
						//		the downloader didn't send any request blocks. Then the connection times out..
						//	Note (ePlus):
						//		ePlus is also using OP_QUEUERANKING to detect end of the session
							Packet	   *pCancelTransferPacket = new Packet(OP_OUTOFPARTREQS, 0);

							g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pCancelTransferPacket->m_dwSize);
							m_pRequestSocket->SendPacket(pCancelTransferPacket,true,true);

							g_eMuleApp.m_pUploadQueue->AddClientToWaitingQueue(this, true);

							return 0;
						}
					}

				//	Feed data to this slot, one block at a time, as long as there's enough bandwidth left
				//	for the entire next block. There must also be blocks to send (not empty request queue)
					while (!m_blockSendQueue.empty())
					{
					//	Get next block and give it to the Request Socket for transfer
						Packet	   *pPacket = m_blockSendQueue.front();
						uint32		dwBlockSize = pPacket->GetRealPacketSize();

						if (dwBlockSize + dwAmountTransferred > dwMaxAmount)
							break;

						m_blockSendQueue.pop_front();

					//	Extended statistics information based on which client software and which port we sent this data to.
					//	This also updates the grand total for sent bytes, etc. And where this data came from.
					//	We have to call function before packet will be sent, otherwise packet will be deleted.
						g_eMuleApp.m_pGlobPrefs->Add2SessionTransferData( GetClientSoft(),
																		GetUserPort(), pPacket->IsFromPF(),
																		TRANSFERTYPE_UPLOAD,
																		dwBlockSize,
																		IsCommunity(),
																		pPacket->GetFilePriority() );
						m_pRequestSocket->SendPacket(pPacket, true, false);
						pPacket = NULL;

					//	We've transferred some data. Store this in relevant statistics variables
						m_dwTransferredUp += dwBlockSize;
						dwAmountTransferred += dwBlockSize;
						m_pCredits->AddUploaded(dwBlockSize, this->GetIP());

					//	If we're out of prepared blocks, create some more
						if (m_blockSendQueue.empty())
							CreateNextBlockPackage();
					}

				//	Check result of the data transfer
					if (dwAmountTransferred == 0)
					{
					//	The size of the packet is larger than allowed bandwidth
						eResult = BSR_FAILED_NOT_ENOUGHT_BANDWIDTH;
					}
					else
					{
						if (!m_blockSendQueue.empty())
						//	If this slot wasn't given enough data to empty its request queue, it wants more data.
							eResult = BSR_OK_WANTS_MORE_BANDWIDTH;
						else
						//	Very rare case, the size of the packets fits the bandwidth
							eResult = BSR_OK;

					//	Remember when this connection was last fed some data, so we
					//	can make sure we give it data often enough to not time out.
						SetLastGotULData();
					}
				}
			}
		}
		else
		{
			AddDebugLogLine(_T("CUpDownClient::SendBlockData (the socket does not exist)"));
			Disconnected(false);
			return 0;
		}
#endif //OLD_SOCKETS_ENABLED
	}
//	In case of exception, disconnect the client
	catch(CException *err)
	{
		g_eMuleApp.m_pdlgEmule->AddBugReport(__FUNCTION__, __FILE__, __LINE__, GetErrorMessage(err));
		err->Delete();
		Disconnected(false);
	}
	catch(...)
	{
		Disconnected(false);
	}

	return dwAmountTransferred;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//	FlushSendBlocks() sends the rest of the last block queued for upload to this client if it has been split
//	across multiple packets. Call this when you stop uploading or the request socket is not able to send.
void CUpDownClient::FlushSendBlocks()
{
	EMULE_TRY

	bool	bBreak = false;

#ifdef OLD_SOCKETS_ENABLED
	while ( !m_blockSendQueue.empty() && m_blockSendQueue.front()->IsSplit() &&
			m_pRequestSocket != NULL && m_pRequestSocket->IsConnected() && !bBreak )
	{
		Packet	   *pPacket = m_blockSendQueue.front();

		m_blockSendQueue.pop_front();

		bool	bBreak = pPacket->IsLastSplit();

		g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
		m_pRequestSocket->SendPacket(pPacket, true, false);
	}
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendHashsetPacket(char *pcFileHash)
{
	EMULE_TRY

	CKnownFile		*pKnownFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(reinterpret_cast<BYTE*>(pcFileHash));

	if (pKnownFile == NULL)
	{
		AddDebugLogLine(RGB_LOG_ERROR + _T("Client '%s' (%s): requested file not found (hash: %s)"), GetUserName(), GetClientNameAndVersionString(), HashToString(reinterpret_cast<BYTE*>(pcFileHash)));
	//	Cause OP_HASHSETREQUEST can be abused for file-scanning
		UpdateFailedFileRequests();
		return;
	}

	uint16				uPartCount = pKnownFile->GetHashCount();
	uint32				dwHashSetLength = 16 * uPartCount;
	CSafeMemFile		pbyteFileData(16 + 2 + dwHashSetLength);

	pbyteFileData.Write(pKnownFile->GetFileHash(), 16);
	pbyteFileData.Write(&uPartCount, 2);

	if (uPartCount != 0 && pKnownFile->GetHashSet() != NULL)
		pbyteFileData.Write(pKnownFile->GetHashSet(), dwHashSetLength);

	Packet	   *pPacket = new Packet(&pbyteFileData);

	pPacket->m_eOpcode = OP_HASHSETANSWER;
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadFileRequest(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED

	m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FlushSendBlocks() was removed, since the client on the other side ignores all packets,
// that will be coming after an OP_CANCEL... what actually means traffic waste.
void CUpDownClient::ClearUploadBlockRequests()
{
	EMULE_TRY

	while (!m_blockRequestsQueue.empty())
	{
		delete m_blockRequestsQueue.front();
		m_blockRequestsQueue.pop_front();
	}

	while (!m_blockSendQueue.empty())
	{
		delete m_blockSendQueue.front();
		m_blockSendQueue.pop_front();
	}

	while (!m_doneBlocksList.empty())
	{
		delete m_doneBlocksList.front();
		m_doneBlocksList.pop_front();
	}

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendRankingInfo()
{
	EMULE_TRY

#ifdef OLD_SOCKETS_ENABLED
	if (m_pRequestSocket == NULL)
		return;

#endif //OLD_SOCKETS_ENABLED
	if (!ExtProtocolAvailable())
		return;

	uint16 uRank = g_eMuleApp.m_pUploadQueue->GetWaitingPosition(this);

	if (uRank == 0)
		return;

	Packet	   *pPacket = new Packet(OP_QUEUERANKING, 12, OP_EMULEPROT);

	memset(pPacket->m_pcBuffer, 0, 12);
	*reinterpret_cast<uint16 *>(pPacket->m_pcBuffer) = uRank;

	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
	m_pRequestSocket->SendPacket(pPacket, true, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SendCommentInfo(CKnownFile *pKnownFile)
{
	EMULE_TRY

	if (!m_bCommentDirty || pKnownFile == NULL || !ExtProtocolAvailable() || m_byteAcceptCommentVer < 1)
		return;

	m_bCommentDirty = false;

	EnumPartFileRating	eRating = pKnownFile->GetFileRating();
	CString				strDesc = pKnownFile->GetFileComment();

	if (eRating == PF_RATING_NONE && strDesc.IsEmpty())
		return;

	CSafeMemFile		packetStream(256);

	packetStream.Write(&eRating, sizeof(eRating));			// <rating 1>

	uint32		dwLength = strDesc.GetLength();

	if (dwLength > 128)
		dwLength = 128;

	packetStream.Write(&dwLength, sizeof(dwLength));		// <commentlen 4>

	if (dwLength > 0)
		packetStream.Write(strDesc.GetString(), dwLength);	// (<commentchars 1>)*commentlen

	Packet	   *pPacket = new Packet(&packetStream, OP_EMULEPROT);

	pPacket->m_eOpcode = OP_FILEDESC;
	g_eMuleApp.m_pUploadQueue->AddUpDataOverheadOther(pPacket->m_dwSize);
#ifdef OLD_SOCKETS_ENABLED
	m_pRequestSocket->SendPacket(pPacket, true);
#endif //OLD_SOCKETS_ENABLED

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::AddRequestCount(uchar *pbyteFileId)
{
	EMULE_TRY

	for (POSITION pos = m_requestedFilesList.GetHeadPosition(); pos != NULL;)
	{
		Requested_File_Struct	   *pReqFile = static_cast<Requested_File_Struct*>(m_requestedFilesList.GetNext(pos));

		if (!md4cmp(pReqFile->m_fileHash, pbyteFileId))
		{
			if (::GetTickCount() - pReqFile->m_dwLastAskedTime < MIN_REQUESTTIME)
			{
				if (GetDownloadState() != DS_DOWNLOADING)
					pReqFile->m_byteNumBadRequests++;
				if (pReqFile->m_byteNumBadRequests == BADCLIENTBAN)
					Ban( _T("AUTOMATED MESSAGE: You're being banned because your client is behaving aggressively!"),
						 GetResString(IDS_BAN_AGGRESSIVECLI) );
			}
			else
			{
				if (pReqFile->m_byteNumBadRequests)
					pReqFile->m_byteNumBadRequests--;
			}
			pReqFile->m_dwLastAskedTime = ::GetTickCount();
			return;
		}
	}

//	If file was not found create a new file
	Requested_File_Struct	   *pNewFile = new Requested_File_Struct;

	memzero(pNewFile, sizeof(Requested_File_Struct));
	md4cpy(pNewFile->m_fileHash, pbyteFileId);
	pNewFile->m_dwLastAskedTime = ::GetTickCount();
	m_requestedFilesList.AddHead(pNewFile);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::UnBan()
{
	EMULE_TRY

	m_bBanned = false;
	m_dwBanTime = 0;

	SetWaitStartTime();
	g_eMuleApp.m_pUploadQueue->UpdateBanCount();
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();

	const uint64 qwIPPortID = GetUID();

	g_eMuleApp.m_pClientList->m_mapOffenseCounter.erase(qwIPPortID);
	g_eMuleApp.m_pClientList->m_mapHashBase.erase(qwIPPortID);

	for (POSITION pos = m_requestedFilesList.GetHeadPosition(); pos != NULL;)
	{
		Requested_File_Struct	   *pReqFile = static_cast<Requested_File_Struct*>(m_requestedFilesList.GetNext(pos));

		pReqFile->m_byteNumBadRequests = 0;
		pReqFile->m_dwLastAskedTime = 0;
	}
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::Ban(LPCTSTR pszLine, const CString &strReason, ...)
{
	EMULE_TRY

	if (!g_eMuleApp.m_pGlobPrefs->BanEnabled())
	{
	//	If the ban function is disabled, no need to ban a client (default is enabled!)
		m_bBanned = false;
		m_dwBanTime = 0;
		return;
	}

	if (m_eUploadState == US_UPLOADING)
	{
	//	Kick already downloading client back to waitingqueue
		AddDebugLogLine(RGB_LOG_DIMMED + _T("Client '%s' (%s) banned and kicked out of uploadqueue %s"), GetUserName(), GetClientNameAndVersionString(), strReason);
		g_eMuleApp.m_pUploadQueue->RemoveFromUploadQueue(this, ETS_BAN, true);
		g_eMuleApp.m_pUploadQueue->AddClientToWaitingQueue(this, true);
	}

	if (m_bBanned)
	{
	//	No need to ban a client (and send a message) again except for extending the bantime
		m_dwBanTime = ::GetTickCount();
		return;
	}

	if (g_eMuleApp.m_pGlobPrefs->CommunityNoBanEnabled() && g_eMuleApp.m_pGlobPrefs->CommunityEnabled() && IsCommunity())
		return;

	m_bBanned = true;

	g_eMuleApp.m_pUploadQueue->UpdateBanCount();
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.UpdateUploadHeader();

	m_dwBanTime = ::GetTickCount();

	uint32	dwResID = IDS_CLIENTBLOCKED;

	if (g_eMuleApp.m_pGlobPrefs->IsBanMessageEnabled())
	{
		CString	strMsg = pszLine;

		strMsg += " Get more info at http://emuleplus.info/ban/";
		g_eMuleApp.m_pdlgEmule->m_wndChat.m_ctlChatSelector.StartSession(this, false);
		g_eMuleApp.m_pdlgEmule->m_wndChat.m_ctlChatSelector.SendMessage(strMsg);
		g_eMuleApp.m_pdlgEmule->m_wndChat.m_ctlChatSelector.EndSession(this, false);
		dwResID = IDS_SENTBAN;
	}

	if (!g_eMuleApp.m_pGlobPrefs->IsCMNotLog())
		AddLogLine(false, RGB_LOG_DIMMED + GetResString(dwResID), GetUserName(), strReason);

	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlQueueList.UpdateClient(this);

	EMULE_CATCH
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
CString CUpDownClient::GetDownloadStateAsString()
{
	return GetStatusDLQueueString((EnumDLQState)GetDownloadState());
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint16	CUpDownClient::GetRemoteRatio()
{
	EMULE_TRY

//	This calculation is not very accurate for several reasons:
//	- We do not know the file priority
//	- We do not save the client info if he has nothing uploaded to us
//	- Other factores like community etc

	double dblBaseValue = 100.0;

#ifdef OLD_SOCKETS_ENABLED
	if (g_eMuleApp.m_pServerConnect->IsLowID())
		dblBaseValue *= 0.8;

#endif //OLD_SOCKETS_ENABLED
	CKnownFile	   *pKnownReqFile = g_eMuleApp.m_pSharedFilesList->GetFileByID(m_reqFileHash);

	dblBaseValue *= GetRemoteBaseModifier();
//	Official client doesn't have SF/RF push
	if ((pKnownReqFile != NULL) && (GetClientSoft() == SO_PLUS))
	{
		dblBaseValue *= pKnownReqFile->GetSizeRatio();
		dblBaseValue *= pKnownReqFile->GetPopularityRatio();
	}
	return static_cast<uint32>(dblBaseValue);

	EMULE_CATCH

	return 100;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
double CUpDownClient::GetRemoteBaseModifier()
{
	EMULE_TRY

//	This calculation is not very accurate for some reasons...
//	Crashes can cause that either client can not write its client data.
//	This can cause differences of the client data at the two clients.
	if (m_pCredits == reinterpret_cast<CClientCredits*>(-1))
		m_pCredits = NULL;

	if (m_pCredits != NULL)
		return m_pCredits->GetScoreRatio(this->GetIP(), true);

	EMULE_CATCH

	return 1.0;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
uint32 CUpDownClient::GetCurrentlyUploadingPart()
{
	EMULE_TRY

	if (!m_doneBlocksList.empty())
	{
	//	Do we need to check upload state here?
		Requested_Block_Struct		*pFirstDoneBlock = m_doneBlocksList.front();

	//	Calculate which part the block is in.
		uint32		dwPartNum = pFirstDoneBlock->m_dwStartOffset / PARTSIZE;

		return dwPartNum;
	}
	else
	{
		return 0xFFFF;
	}

	EMULE_CATCH

	return 0xFFFF;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void CUpDownClient::SetUploadState(EnumULQState eUploadState)
{
	m_eUploadState = eUploadState;
	g_eMuleApp.m_pdlgEmule->m_wndTransfer.m_ctlClientList.UpdateClient(this);
}
